#! /usr/bin/env python
# encoding: utf-8

import os,sys
import Configure,Options,Utils
import ccroot,ar
from Configure import conftest
def find_gxx(conf):
	cxx=conf.find_program(['g++','c++'],var='CXX',mandatory=True)
	cxx=conf.cmd_to_list(cxx)
	ccroot.get_cc_version(conf,cxx,gcc=True)
	conf.env.CXX_NAME='gcc'
	conf.env.CXX=cxx
def gxx_common_flags(conf):
	v=conf.env
	v['CXXFLAGS_DEBUG']=['-g']
	v['CXXFLAGS_RELEASE']=['-O2']
	v['CXX_SRC_F']=''
	v['CXX_TGT_F']=['-c','-o','']
	v['CPPPATH_ST']='-I%s'
	if not v['LINK_CXX']:v['LINK_CXX']=v['CXX']
	v['CXXLNK_SRC_F']=''
	v['CXXLNK_TGT_F']=['-o','']
	v['LIB_ST']='-l%s'
	v['LIBPATH_ST']='-L%s'
	v['STATICLIB_ST']='-l%s'
	v['STATICLIBPATH_ST']='-L%s'
	v['RPATH_ST']='-Wl,-rpath,%s'
	v['CXXDEFINES_ST']='-D%s'
	v['SONAME_ST']='-Wl,-h,%s'
	v['SHLIB_MARKER']='-Wl,-Bdynamic'
	v['STATICLIB_MARKER']='-Wl,-Bstatic'
	v['FULLSTATIC_MARKER']='-static'
	v['program_PATTERN']='%s'
	v['shlib_CXXFLAGS']=['-fPIC','-DPIC']
	v['shlib_LINKFLAGS']=['-shared']
	v['shlib_PATTERN']='lib%s.so'
	v['staticlib_LINKFLAGS']=['-Wl,-Bstatic']
	v['staticlib_PATTERN']='lib%s.a'
	v['LINKFLAGS_MACBUNDLE']=['-bundle','-undefined','dynamic_lookup']
	v['CCFLAGS_MACBUNDLE']=['-fPIC']
	v['macbundle_PATTERN']='%s.bundle'
def gxx_modifier_win32(conf):
	v=conf.env
	v['program_PATTERN']='%s.exe'
	v['shlib_PATTERN']='%s.dll'
	v['implib_PATTERN']='lib%s.dll.a'
	v['IMPLIB_ST']='-Wl,--out-implib,%s'
	dest_arch=v['DEST_CPU']
	v['shlib_CXXFLAGS']=[]
	v.append_value('shlib_CXXFLAGS','-DDLL_EXPORT')
	v.append_value('LINKFLAGS','-Wl,--enable-auto-import')
def gxx_modifier_cygwin(conf):
	gxx_modifier_win32(conf)
	v=conf.env
	v['shlib_PATTERN']='cyg%s.dll'
	v.append_value('shlib_LINKFLAGS','-Wl,--enable-auto-image-base')
def gxx_modifier_darwin(conf):
	v=conf.env
	v['shlib_CXXFLAGS']=['-fPIC','-compatibility_version','1','-current_version','1']
	v['shlib_LINKFLAGS']=['-dynamiclib']
	v['shlib_PATTERN']='lib%s.dylib'
	v['staticlib_LINKFLAGS']=[]
	v['SHLIB_MARKER']=''
	v['STATICLIB_MARKER']=''
	v['SONAME_ST']=''
def gxx_modifier_aix(conf):
	v=conf.env
	v['program_LINKFLAGS']=['-Wl,-brtl']
	v['shlib_LINKFLAGS']=['-shared','-Wl,-brtl,-bexpfull']
	v['SHLIB_MARKER']=''
def gxx_modifier_platform(conf):
	dest_os=conf.env['DEST_OS']or Utils.unversioned_sys_platform()
	gxx_modifier_func=globals().get('gxx_modifier_'+dest_os)
	if gxx_modifier_func:
		gxx_modifier_func(conf)
def detect(conf):
	conf.find_gxx()
	conf.find_cpp()
	conf.find_ar()
	conf.gxx_common_flags()
	conf.gxx_modifier_platform()
	conf.cxx_load_tools()
	conf.cxx_add_flags()

conftest(find_gxx)
conftest(gxx_common_flags)
conftest(gxx_modifier_win32)
conftest(gxx_modifier_cygwin)
conftest(gxx_modifier_darwin)
conftest(gxx_modifier_aix)
conftest(gxx_modifier_platform)



from __future__ import division
import contextlib, math, logging
import numpy as np
import numpy.core.arrayprint as npap

class CircuitException(Exception):
    def __init__(self, msg, logger=None):
        Exception.__init__(self, msg)
        self.logger = logger

# http://stackoverflow.com/questions/2891790/pretty-printing-of-numpy-array
@contextlib.contextmanager
def printoptions(strip_zeros=True, **kwargs):
    """Context manager to apply Numpy printoptions locally.

    Example usage:
    >>> x = np.array([0.078, 0.480, 0.413, 0.830, 0.776])
    >>> with printoptions(precision=3, suppress=True, strip_zeros=False):
            print(x)

    """
    origcall = npap.FloatFormat.__call__
    def __call__(self, x, strip_zeros=strip_zeros):
        return origcall.__call__(self, x, strip_zeros)
    npap.FloatFormat.__call__ = __call__
    original = np.get_printoptions()
    np.set_printoptions(**kwargs)
    yield
    np.set_printoptions(**original)
    npap.FloatFormat.__call__ = origcall

def pow2roundup(x):
    "return smallest 2**Y which is >= x (with x < 2**32)"
    if x <= 0:
        return 0
    x -= 1
    x |= x >> 1;
    x |= x >> 2;
    x |= x >> 4;
    x |= x >> 8;
    x |= x >> 16;
    return x+1;

def genlogsweep(fmin, fmax, rate, k0, k1, k2, dtype=np.float64):
    """generate logarithmic sweep signal
    
    fmin: start frequency
    fmax: end frequency
    rate: sample rate
    k0: fade in before start of signal (samples)
    k1: length of signal (samples)
    k2: fade out after end of signal (samples)
    dtype: data type of signal

    returns: (logsweep signal, inverse logsweep signal)
    """
    s1 = np.empty(k0 + k1 + k2, dtype=dtype)
    s2 = np.empty_like(s1)
    b = math.log(fmax / fmin) / k1
    a = fmin / (b * rate)
    q0 = a * math.exp(-b * k0)
    def sweep(t, g=None):
        q = a * np.exp(b * t)
        p = q - q0
        x = np.sin(2 * np.pi * (p - np.floor(p)))
        if g is not None:
            x *= g
        return x, x * q
    if k0:
        t = np.arange(-k0,0)
        s1[:k0], s2[:k0] = sweep(t, np.cos(0.5 * np.pi * t / k0))
    t = np.arange(k1)
    s1[k0:k0+k1], s2[k0:k0+k1] = sweep(t)
    if k2:
        t = np.arange(k1,k1+k2)
        s1[k0+k1:], s2[k0+k1:] = sweep(t, np.sin(0.5 * np.pi * (k1 + k2 - t) / k2))
    s2 = s2[::-1] * 4 * b * b
    return s1, s2, fmin, fmax, k1 / math.log(fmax / fmin)

def expchirp(fmin, fmax, rate, k0, k1, k2, dtype=np.float64):
    nyq = rate / 2
    if fmax <= fmin:
        fmax = nyq
    if k1 <= 0:
        k1 = 1
    p = np.ceil(np.log2(nyq / fmin)) + 1
    #p1 = np.floor(np.log2(nyq / fmax))
    p1 = 0
    ep = 2 ** p
    fmin = nyq / ep
    fmax = nyq / (2 ** p1)
    lep = np.log(ep)
    L1 = 2 * ep * lep
    L = np.ceil(k1 / L1) * L1
    N = np.ceil(L)
    NN = np.ceil(L - p1 * N / p)
    epN = 2 ** (p/N)
    n = np.arange(NN)
    pp = L / L1 * epN**n
    s = np.sin(2 * np.pi * pp)
    #if k0:
    #    nn = int(round(N/p))
    #    s[:nn] = s[:nn] * np.hanning(2*nn)[:nn]
    si = 2/N * lep * s[::-1] / epN**n
    if p1:
        si *= 2 ** -p1
    return s, si, fmin, fmax, N / lep

def fft_convolve(h, xd, invert=False):
    n = len(h) + len(xd) - 1
    n2 = pow2roundup(n)
    H = np.fft.rfft(h[2:], n2, axis=0)
    XD = np.fft.rfft(xd, n2, axis=0)
    if len(XD.shape) == 2 and len(H.shape) == 1:
        H = H.reshape(len(H), 1)
    if invert:
        XD /= H
    else:
        XD *= H
    s = np.fft.irfft(XD, n2, axis=0)
    return s[:n]

def calc_grid(func, grd, nvals):
    grd_shape = grd.shape
    numpoints = np.product(grd_shape[1:])
    grd = grd.reshape(grd_shape[0], numpoints)
    fnc = func(grd.T).T
    return fnc.reshape((nvals,)+grd_shape[1:])

def mkgrid(axeslist):
    if len(axeslist) > 1:
        grd = np.array(np.meshgrid(*axeslist, indexing='ij'))
    else:
        grd = np.empty((1, len(axeslist[0])))
        grd[0] = axeslist[0]
    return grd

def error(text, logger=None):
    raise CircuitException(text, logger)



import os
from libc.string cimport const_char, strcmp
from libc.stdlib cimport malloc, free
import numpy as np
cimport numpy as np
np.import_array()
cdef extern from "time.h":
    cdef struct timespec:
        int tv_sec
        int tv_nsec
    ctypedef int clockid_t
    int CLOCK_MONOTONIC
    int clock_gettime(clockid_t clk_id, timespec *tp)

cdef extern from "circuit.hpp":
    ctypedef struct N_Vector:
        pass
    N_Vector N_VNew_Serial(long int vec_length)
    void N_VDestroy_Serial(N_Vector v)
    double *NV_DATA_S(N_Vector v)
    int NV_LENGTH_S(N_Vector v)

    int FS "fs"

    cppclass UserData "ComponentBase::UserData":
        double *inval
        double *state

    cppclass ComponentBase:
        int NEQ
        int NDIM
        int NVALS
        int N_IN
        int N_OUT
        int n_params
        double *params
        char **param_names
        char **in_names
        char **out_names
        char **state_names
        char **var_names
        int *ix
        int verbose
        int func(N_Vector x, N_Vector u, UserData *user_data)
        void update(N_Vector y, N_Vector x, N_Vector state)
        int startvalue(N_Vector x, N_Vector s, N_Vector u)
        int findzero(double *x, double *s, N_Vector u) 
        int set_range(int i, double lower, double upper, int size)
        void init(N_Vector u0, char* fname)
        int calc(N_Vector x, N_Vector s, N_Vector u)

    cppclass TriodeCircuit(ComponentBase):
        pass

    cppclass CoupledTriodeCircuit(ComponentBase):
        pass

    cppclass PowerAmpGate(ComponentBase):
        pass

    cppclass PowerAmpPlate(ComponentBase):
        pass

    cppclass PhaseSplitter(ComponentBase):
        void feedback_coeff(double *b0, double *b1, double *a1, double pres)

    cdef cppclass CSpline "Spline":
        CSpline(double *, int, double, double)
        double eval(double)

fs = FS


class ConvergenceError(ValueError):

    def __init__(self, err, x, s, tag=None):
        self.error = err
        self.x = x
        self.s = s
        self.tag = tag
        t = "%s: " % tag if tag else ""
        v1 = ", ".join((str(v) for v in x))
        if s is not None:
            v2 = "/[%s]" % ", ".join((str(v) for v in s))
        else:
            v2 = ""
        ValueError.__init__(
            self, "%sKINSol error %d at [%s]%s" % (t, err, v1, v2))

    def __reduce__(self):
        return (ConvergenceError, (self.error, self.x, self.s, self.tag))


cdef int find_idx(const_char *k, const_char **p, int n):
    cdef int i
    for i in range(n):
        if strcmp(p[i], k) == 0:
            return i
    return -1

cdef to_list(const_char **p, int n):
    cdef int i
    cdef list l = []
    cdef char *t
    for i in range(n):
        t = <char*>(p[i])
        l.append(t)
    return l

cdef inline double ts_diff(timespec ts1, timespec ts2):
    cdef double df = ts1.tv_sec - ts2.tv_sec
    return df * 1e9 + (ts1.tv_nsec - ts2.tv_nsec)

cdef class CalcBase:
    cdef ComponentBase *pbase
    cdef int *idx
    cdef int *capt_idx
    cdef int capt_len
    cdef list capt_vars
    cdef np.ndarray _captured
    cdef N_Vector _state
    cdef np.ndarray state_min
    cdef np.ndarray state_max
    cdef N_Vector inp
    cdef N_Vector outp
    cdef double time_per_sample

    property NDIM:
        def __get__(self):
            return self.pbase.NDIM
    property NEQ:
        def __get__(self):
            return self.pbase.NEQ
    property NVALS:
        def __get__(self):
            return self.pbase.NVALS
    property N_IN:
        def __get__(self):
            return self.pbase.N_IN
    property N_OUT:
        def __get__(self):
            return self.pbase.N_OUT
    property param_names:
        def __get__(self):
            return to_list(self.pbase.param_names, self.pbase.n_params)
    property in_names:
        def __get__(self):
            return to_list(self.pbase.in_names, self.pbase.N_IN)
    property out_names:
        def __get__(self):
            return to_list(self.pbase.out_names, self.pbase.N_OUT)
    property state_names:
        def __get__(self):
            return to_list(self.pbase.state_names, self.pbase.NDIM-self.pbase.N_IN)
    property var_names:
        def __get__(self):
            return to_list(self.pbase.var_names, self.pbase.NEQ)
    property state:
        def __get__(self):
            cdef int i
            return np.array([NV_DATA_S(self._state)[i] for i in range(self.pbase.NDIM-self.pbase.N_IN)])
        def __set__(self, v):
            for i in range(self.pbase.NDIM-self.pbase.N_IN):
                NV_DATA_S(self._state)[i] = v[i]
    property min_state:
        def __get__(self):
            return self.state_min
    property max_state:
        def __get__(self):
            return self.state_max
    property nanosec_per_sample:
        "time in nanoseconds measured for last compute call"
        def __get__(self):
            return self.time_per_sample

            
    def __dealloc__(self):
        free(self.idx)
        free(self.capt_idx)
        N_VDestroy_Serial(self.inp)
        N_VDestroy_Serial(self._state)
        N_VDestroy_Serial(self.outp)
        del self.pbase

    property capture_signals:
        def __get__(self):
            return self.capt_vars
        def __set__(self, v):
            cdef int i, j
            cdef char *t
            if isinstance(v, basestring):
                v = [v]
            l = []
            for n in v:
                t = n
                j = find_idx(t, self.pbase.var_names, self.pbase.NEQ)
                if j < 0:
                    raise ValueError("not found: %s", n)
                l.append(j)
            i = len(l)
            if i != self.capt_len:
                free(self.capt_idx)
                self.capt_idx = <int*>malloc(i * sizeof(int))
                self.capt_len = i
            for j in range(i):
                self.capt_idx[j] = l[j]
            self.capt_vars = v

    property captured:
        def __get__(self):
            return self._captured

    cdef int set_var(self, char *k, double v):
        for i in range(self.pbase.n_params):
            if strcmp(k, self.pbase.param_names[i]) == 0:
                self.pbase.params[i] = v
                return 1
        return 0           

    def init(self):
        cdef int i
        for k, v in self.circuit.items():
            if not self.set_var(k, v):
                raise KeyError("unknown: %s" % k)
        for i, l in enumerate(self.start_grid):
            if not self.pbase.set_range(i, l[0], l[1], l[2]):
                raise ValueError("can't set range nr %d" % i)
        if len(self.u0) != self.pbase.NEQ:
            raise ValueError("lenght %d expected for u0" % self.pbase.NEQ)
        for i, v in enumerate(self.u0):
            NV_DATA_S(self.outp)[i] = v
        if "AMPSIM_CACHE" in os.environ:
            fname = "%s.cache" % self.comp_id
        else:
            fname = None
        if fname and os.path.exists(fname) and os.stat(fname).st_mtime < os.stat("circuit/components.py").st_mtime:
            print "remove", fname
            os.remove(fname)
        cdef char *_fname = NULL
        if fname:
            _fname = fname
        self.pbase.init(self.outp, _fname)

    def pre_call(self, A, with_state):
        return A

    def post_call(self, O, with_state):
        return O

    cdef np.ndarray prepare_input(self, object a, int with_state, int* n_in, int* n_out, int* ndim):
        cdef np.ndarray A = np.PyArray_FROMANY(a, np.NPY_DOUBLE, 0, 2, np.NPY_ALIGNED)
        ndim[0] = np.PyArray_NDIM(A)
        n_in[0] = self.pbase.NDIM if with_state else self.pbase.N_IN
        n_out[0] = self.pbase.N_OUT
        if with_state:
            n_out[0] += self.pbase.NDIM-self.pbase.N_IN
        if ndim[0] == 0:
            if n_in[0] > 1:
                raise ValueError("input array: bad shape")
            A = np.PyArray_Reshape(A, (1, 1))
            if n_out[0] > 1:
                ndim[0] = 1
        elif ndim[0] == 1:
            if n_in[0] == 1:
                A = np.PyArray_Reshape(A, ((np.PyArray_DIM(A, 0), 1)))
                if n_out[0] > 1:
                    ndim[0] = 2
            elif np.PyArray_DIM(A, 0) == n_in[0]:
                A = np.PyArray_Reshape(A, (1, n_in[0]))
                if n_out[0] == 1:
                    ndim[0] = 0
            else:
                raise ValueError("input array: bad shape")
        elif np.PyArray_NDIM(A) == 2:
            if np.PyArray_DIM(A, 1) != n_in[0]:
                raise ValueError("input array: bad shape %d/%d" % (np.PyArray_DIM(A, 1), n_in[0]))
            if n_out[0] == 1:
                ndim[0] = 1
        return A

    def __call__(self, a, int with_state=False):
        cdef int ndim = 0
        cdef int n_in = 0
        cdef int n_out = 0
        cdef np.ndarray A = self.prepare_input(a, with_state, &n_in, &n_out, &ndim)
        cdef np.npy_intp dim[2]
        dim[0] = np.PyArray_DIM(A, 0)
        dim[1] = n_out
        cdef np.ndarray O = np.PyArray_SimpleNew(2, dim, np.NPY_DOUBLE)
        cdef np.flatiter itc
        if self.capt_len:
            dim[0] = np.PyArray_DIM(A, 0)
            dim[1] = self.capt_len
            self._captured = np.PyArray_SimpleNew(2, dim, np.NPY_DOUBLE)
            itc = np.PyArray_IterNew(self._captured)
        A = self.pre_call(A, with_state)
        cdef int i, j, n
        cdef np.flatiter it = np.PyArray_IterNew(A)
        cdef np.flatiter ito = np.PyArray_IterNew(O)
        cdef np.npy_intp ix[2]
        cdef timespec t0, t1
        cdef double v
        clock_gettime(CLOCK_MONOTONIC, &t0)
        for i in range(np.PyArray_DIM(A, 0)):
            ix[0] = i
            for j in range(self.pbase.N_IN):
                ix[1] = j
                np.PyArray_ITER_GOTO(it, ix)
                NV_DATA_S(self.inp)[j] = (<double*>np.PyArray_ITER_DATA(it))[0]
            if with_state:
                for j in range(n_in - self.pbase.N_IN):
                    ix[1] = self.pbase.N_IN + j
                    np.PyArray_ITER_GOTO(it, ix)
                    NV_DATA_S(self._state)[j] = (<double*>np.PyArray_ITER_DATA(it))[0]
            n = self.pbase.calc(self.inp, self._state, self.outp)
            if n:
                raise ConvergenceError(n, A[i], None if with_state else self.state, getattr(self,"comp_id",None))
            if self.capt_len:
                for j in range(self.capt_len):
                    (<double*>np.PyArray_ITER_DATA(itc))[j] = NV_DATA_S(self.outp)[self.capt_idx[j]]
                    np.PyArray_ITER_NEXT(itc)
            self.pbase.update(self.outp, self.inp, self._state)
            for j in range(self.pbase.N_OUT):
                (<double*>np.PyArray_ITER_DATA(ito))[0] = NV_DATA_S(self.outp)[self.idx[j]]
                np.PyArray_ITER_NEXT(ito)
            if with_state:
                for j in range(n_out-self.pbase.N_OUT):
                    (<double*>np.PyArray_ITER_DATA(ito))[0] = NV_DATA_S(self._state)[j]
                    np.PyArray_ITER_NEXT(ito)
            else:
                for j in range(self.pbase.NDIM - self.pbase.N_IN):
                    v = NV_DATA_S(self._state)[j]
                    if v > (<double*>np.PyArray_DATA(self.state_max))[j]:
                        (<double*>np.PyArray_DATA(self.state_max))[j] = v
                    if v < (<double*>np.PyArray_DATA(self.state_min))[j]:
                        (<double*>np.PyArray_DATA(self.state_min))[j] = v
        clock_gettime(CLOCK_MONOTONIC, &t1)
        self.time_per_sample = ts_diff(t1,t0)/np.PyArray_DIM(A, 0)
        O = self.post_call(O, with_state)
        cdef np.PyArray_Dims dm
        if ndim != np.PyArray_NDIM(O):
            if ndim == 1:
                if np.PyArray_NDIM(O) == 0:
                    dim[0] = 1
                else:
                    dim[0] = np.PyArray_DIM(O, 0) * np.PyArray_DIM(O, 1)
            else:
                if np.PyArray_NDIM(O) == 0:
                    dim[0] = 1
                    dim[1] = 1
                else:
                    dim[0] = np.PyArray_DIM(O, 0)
                    dim[1] = 1
            dm.ptr = dim
            dm.len = ndim
            O = np.PyArray_Newshape(O, &dm, np.NPY_ANYORDER)
        return O

    def func(self, a, state):
        cdef int i
        cdef N_Vector x = N_VNew_Serial(self.pbase.N_IN)
        cdef N_Vector s = N_VNew_Serial(self.pbase.NDIM-self.pbase.N_IN)
        cdef N_Vector u = N_VNew_Serial(self.pbase.NEQ)
        for i in range(self.pbase.N_IN):
            NV_DATA_S(x)[i] = a[i]
        for i in range(self.pbase.NDIM-self.pbase.N_IN):
            NV_DATA_S(s)[i] = state[i]
        i = self.pbase.calc(x, s, u)
        if i:
            raise ConvergenceError(i, a, state, getattr(self,"comp_id",None))
        self.pbase.update(u, x, s)
        for i in range(self.pbase.NDIM-self.pbase.N_IN):
            state[i] = NV_DATA_S(s)[i]
        ret = [NV_DATA_S(u)[self.idx[i]] for i in range(self.pbase.N_OUT)]
        N_VDestroy_Serial(x)
        N_VDestroy_Serial(s)
        N_VDestroy_Serial(u)
        return ret

    def startvalue(self, x, s):
        cdef int i
        cdef N_Vector X = N_VNew_Serial(self.pbase.N_IN)
        cdef N_Vector S = N_VNew_Serial(self.pbase.NDIM-self.pbase.N_IN)
        cdef N_Vector U = N_VNew_Serial(self.pbase.NEQ)
        for i in range(self.pbase.N_IN):
            NV_DATA_S(X)[i] = x[i]
        for i in range(self.pbase.NDIM-self.pbase.N_IN):
            NV_DATA_S(S)[i] = s[i]
        self.pbase.startvalue(X, S, U)
        ret = [NV_DATA_S(U)[i] for i in range(self.pbase.NEQ)]
        N_VDestroy_Serial(X)
        N_VDestroy_Serial(S)
        N_VDestroy_Serial(U)
        return ret

    def findzero(self, x, s, u, verbose=False):
        cdef int i
        cdef N_Vector X = N_VNew_Serial(self.pbase.N_IN)
        cdef N_Vector S = N_VNew_Serial(self.pbase.NDIM-self.pbase.N_IN)
        cdef N_Vector U = N_VNew_Serial(self.pbase.NEQ)
        for i in range(self.pbase.N_IN):
            NV_DATA_S(X)[i] = x[i]
        for i in range(self.pbase.NDIM-self.pbase.N_IN):
            NV_DATA_S(S)[i] = s[i]
        for i in range(self.pbase.NEQ):
            NV_DATA_S(U)[i] = u[i]
        self.pbase.verbose = verbose
        i = self.pbase.findzero(NV_DATA_S(X), NV_DATA_S(S), U)
        self.pbase.verbose = False
        if i != 0:
            raise ValueError
        ret = [NV_DATA_S(U)[i] for i in range(self.pbase.NEQ)]
        N_VDestroy_Serial(X)
        N_VDestroy_Serial(S)
        N_VDestroy_Serial(U)
        return ret

    def equations(self, x, s, u):
        cdef UserData D
        cdef N_Vector X = N_VNew_Serial(self.pbase.N_IN)
        cdef N_Vector S = N_VNew_Serial(self.pbase.NDIM-self.pbase.N_IN)
        cdef N_Vector U = N_VNew_Serial(self.pbase.NEQ)
        cdef N_Vector F = N_VNew_Serial(self.pbase.NEQ)
        for i in range(self.pbase.N_IN):
            NV_DATA_S(X)[i] = x[i]
        for i in range(self.pbase.NDIM-self.pbase.N_IN):
            NV_DATA_S(S)[i] = s[i]
        for i in range(self.pbase.NEQ):
            NV_DATA_S(U)[i] = u[i]
        D.inval = NV_DATA_S(X)
        D.state = NV_DATA_S(S)
        self.pbase.func(U, F, &D)
        ret = [NV_DATA_S(F)[i] for i in range(self.pbase.NEQ)]
        N_VDestroy_Serial(X)
        N_VDestroy_Serial(S)
        N_VDestroy_Serial(U)
        N_VDestroy_Serial(F)
        return ret

    def __init__(self):
        cdef np.npy_intp i
        cdef int j
        self.inp = N_VNew_Serial(self.pbase.N_IN)
        self._state = N_VNew_Serial(self.pbase.NDIM-self.pbase.N_IN)
        i = self.pbase.NDIM-self.pbase.N_IN
        self.state_min = np.PyArray_SimpleNew(1, &i, np.NPY_DOUBLE)
        self.state_min[:] = 1e99
        self.state_max = np.PyArray_SimpleNew(1, &i, np.NPY_DOUBLE)
        self.state_max[:] = -1e99
        self.outp = N_VNew_Serial(self.pbase.NEQ)
        self.idx = <int*>malloc(2*sizeof(int))
        for i in range(self.pbase.N_OUT):
            j = find_idx(self.pbase.out_names[i], self.pbase.var_names, self.pbase.NEQ)
            if j < 0:
                raise ValueError("not found: %s", self.pbase.out_names[i])
            self.idx[i] = j
        self.state = self.operating_point


cdef class tcParams(CalcBase):
    def __cinit__(self):
        self.pbase = new TriodeCircuit()


cdef class ctcParams(CalcBase):
    def __cinit__(self):
        self.pbase = new CoupledTriodeCircuit()


cdef class pagParams(CalcBase):
    def __cinit__(self):
        self.pbase = new PowerAmpGate()

cdef class papParams(CalcBase):
    def __cinit__(self):
        self.pbase = new PowerAmpPlate()

    def Xpre_call(self, A, with_state):
        return np.append(A[:,0], A[:,1] + A[:,0], axis=1)

    def post_call(self, O, with_state):
        if with_state:
            return np.append(O[:,:1] - O[:,1:2], O[:,2:], axis=1)
        else:
            return O[:,0] - O[:,1]

    def func(self, a, state):
        # expects [Ug1, Ug1+Ug2]
        # returns Ua1 - Ua2
        cdef int i
        cdef N_Vector x = N_VNew_Serial(self.pbase.N_IN)
        cdef N_Vector s = N_VNew_Serial(self.pbase.NDIM-self.pbase.N_IN)
        cdef N_Vector u = N_VNew_Serial(self.pbase.NEQ)
        for i in range(self.pbase.N_IN):
            NV_DATA_S(x)[i] = a[i]
        for i in range(self.pbase.NDIM-self.pbase.N_IN):
            NV_DATA_S(s)[i] = state[i]
        #NV_DATA_S(x)[1] -= NV_DATA_S(x)[0] #added
        i = self.pbase.calc(x, s, u)
        if i:
            raise ConvergenceError(i, a, state)
        self.pbase.update(u, x, s)
        for i in range(self.pbase.NDIM-self.pbase.N_IN):
            state[i] = NV_DATA_S(s)[i]
        ret = [NV_DATA_S(u)[3] - NV_DATA_S(u)[4]] #changed
        N_VDestroy_Serial(x)
        N_VDestroy_Serial(s)
        N_VDestroy_Serial(u)
        return ret


cdef class psParams(CalcBase):
    def __cinit__(self):
        self.pbase = new PhaseSplitter()

    def feedback_coeff(self, double pres):
        cdef double b0, b1, a1
        (<PhaseSplitter*>self.pbase).feedback_coeff(&b0, &b1, &a1, pres)
        return b0, b1, a1


cdef class Spline:
    cdef CSpline *spl
    cdef np.ndarray y

    def __cinit__(self, np.ndarray[np.float_t,ndim=1] y, double xstart, double h):
        self.y = y
        self.spl = new CSpline(<double*>np.PyArray_DATA(y), len(y), xstart, h)

    def __dealloc__(self):
        del self.spl

    def __call__(self, np.ndarray[np.float_t,ndim=1] x):
        cdef int i
        cdef double v
        out = np.empty_like(x)
        for i in range(len(x)):
            out[i] = self.spl.eval(x[i])
        return out
    
# Local Variables:
# compile-command: "rm -f pycircuit.so; python setup.py build_ext --inplace"
# End:



import sympy as sp
import numpy as np

class VariableAccess(object):

    def __init__(self, name, param, pointer):
        if isinstance(name, (MatrixDefinition,MatrixDeclaration)):
            self.name = name.name
            self.pointer = name.pointer
        else:
            assert isinstance(name, basestring)
            self.name = name
            if pointer is None:
                self.pointer = param
            else:
                self.pointer = pointer
        self.param = param

    @staticmethod
    def _len(s):
        if isinstance(s, slice):
            return s.stop - s.start
        return s[1] - s[0]

    @staticmethod
    def _start(s):
        if isinstance(s, slice):
            return s.start
        return s[0]

    def address_of(self):
        if self.param:
            if self.pointer:
                return "par.%s" % self.name
            else:
                return "&par.%s" % self.name
        else:
            if self.pointer:
                return self.name
            else:
                return "&%s" % self.name

    def __str__(self):
        if self.param:
            if self.pointer:
                return "(*par.%s)" % self.name
            else:
                return "par.%s" % self.name
        else:
            if self.pointer:
                return "*%s" % self.name
            else:
                return self.name


class VectorAccess(VariableAccess):

    def __init__(self, name, block=None, param=False, pointer=None):
        VariableAccess.__init__(self, name, param, pointer)
        self.block = block

    @staticmethod
    def block_expr(block):
        start = VariableAccess._start(block)
        if start == 0:
            return ".head<%d>()" % VariableAccess._len(block)
        else:
            return ".segment<%d>(%d)" % (VariableAccess._len(block), start)

    def __str__(self):
        s = VariableAccess.__str__(self)
        if self.block:
            s += self.block_expr(self.block)
        return s


class MatrixAccess(VariableAccess):

    def __init__(self, name, block=None, param=False, pointer=None):
        VariableAccess.__init__(self, name, param, pointer)
        self.block = block

    def __str__(self):
        s = VariableAccess.__str__(self)
        if self.block:
            r = self.block[0]
            c = self.block[1]
            s += ".block<%d,%d>(%d,%d)" % (self._len(r), self._len(c), self._start(r), self._start(c))
        return s


class MatrixDeclaration(object):

    def __init__(self, name, rows, cols=1, mapping=False, pointer=False, aligned=True, array=False):
        self.name = name
        self.rows = rows
        self.cols = cols
        self.mapping = mapping
        self.pointer = pointer
        self.aligned = aligned
        self.array = array

    def generate(self, postfix=""):
        n = self.name
        if postfix:
            n += postfix
        if self.pointer:
            n = "*" + n
        m = "Matrix"
        if self.array:
            m = "Array"
        attr = " "
        if self.mapping:
            if self.aligned:
                attr = ", Aligned"
            return "Map<%s<creal, %d, %d>%s>%s" % (m, self.rows, self.cols, attr, n)
        else:
            return "%s<creal, %d, %d> %s" % (m, self.rows, self.cols, n)

    def generate_lines(self, indent=""):
        return "%s%s;\n" % (indent, self.generate())

    def __str__(self):
        return self.generate()


class MatrixDefinition(object):

    templ_def = "%(prefix)s%(tp)s<creal, %(rows)d, %(cols)d> %(definition)s;\n"

    templ_data = "%(indent)sstatic const creal __attribute__((aligned(16))) %(data_name)s[%(rows)d*%(cols)d] = { %(data)s };\n"

    templ_init = "%(prefix)s%(const)sMap<%(const)s%(tp)s<creal, %(rows)d, %(cols)d>%(attr)s> %(definition)s;\n"

    def __init__(self, name, value=None, rows=None, cols=None, pointer=False, aligned=True, const=None, array=False):
        if isinstance(name, VariableAccess):
            name = name.name
        try:
            s = value.shape
            if len(s) == 1:
                s = (s[0], 1)
            if rows is not None or cols is not None:
                if s != (rows, cols):
                    raise ValueError("inconsistent shape for matrix %s: %s != %s" % (name, s, (rows, cols)))
            rows, cols = s
            if not isinstance(value, sp.Matrix):
                value = sp.Matrix(value)
            self.value_is_matrix = True
        except AttributeError:
            self.value_is_matrix = False
        if rows is None or cols is None:
            raise ValueError("rows and cols must be set or matrix value supplied")
        self.name = name
        self.value = value
        self.rows = rows
        self.cols = cols
        self.pointer = pointer
        self.aligned = aligned
        self.const = const
        self.array = array

    def generate(self, indent=""):
        l = []
        n = self.name
        tp = "Matrix"
        if self.array:
            tp = "Array"
        if self.pointer:
            n = "*" + n
        if self.value is not None:
            const = ""
            if self.value_is_matrix:
                aligned = True
                if self.const is None or self.const:
                    const = "const "
                static = "static "
                data_name = self.name.lower() + "_data"
                l.append(self.templ_data % dict(
                    indent = indent,
                    data_name = data_name,
                    tp = tp,
                    rows = self.rows,
                    cols = self.cols,
                    data = ",".join([sp.ccode(c) for c in self.value.T]), # col-major order is eigen3 default
                    ))
            else:
                aligned = self.aligned
                if self.const:
                    const = "const "
                static = ""
                data_name = self.value
            attr = " "
            if self.aligned:
                attr = ", Aligned"
            l.append(self.templ_init % dict(
                prefix = "%s%s" % (indent, static),
                tp = tp,
                rows = self.rows,
                cols = self.cols,
                attr = attr,
                const = const,
                definition = "%s(%s)" % (n, data_name),
                ))
        else:
            l.append(self.templ_def % dict(
                prefix = indent,
                tp = tp,
                rows = self.rows,
                cols = self.cols,
                definition = n,
                ))
        return "".join(l)

    generate_lines = generate

    def __str__(self):
        return self.generate()


class Namespace(object):

    def __init__(self, indent=""):
        self.indent = indent
        self.ns = {}
        self.generated = set()

    def add(self, m):
        if m.name in self.ns:
            mm = self.ns[m.name]
            if m.rows != mm.rows or m.cols != mm.cols:
                raise ValueError("inkonsistent definition of %s" % m)
            return
        self.ns[m.name] = m

    def generate_lines(self):
        l = []
        for n, m in sorted(self.ns.items()):
            if m in self.generated:
                continue
            l.append(m.generate_lines(self.indent))
            self.generated.add(m)
        return l

    def __str__(self):
        return "".join(self.generate_lines())


def expr_list_to_ccode(ret, ex, idx='[%d]', off=0):
    """
    ret: name of variable
    ex: vector expression (iterable)
    idx: syntax for index access
    """
    ret += idx
    l = []
    for i, ee in enumerate(ex):
        l += [v+"\n" for v in sp.ccode(ee, ret % (i+off)).split("\n")]
    return l

def make_symbol_vector(s, n, idx="[%d]"):
    return sp.symbols(['%s%s' % (s, idx % i) for i in range(n)])

def matrix_is_zero(m):
    return np.allclose(m, np.zeros_like(m))

def matrix_is_identity(m):
    if m.shape[0] != m.shape[1]:
        return False
    return np.allclose(m, np.identity(m.shape[0]))

def join_with_indent(lines, indent="    "):
    return "".join([l+indent for l in lines])

def gen_linear_combination(ns, resname, varname, mult, mult_data, add=None, add_data=None, cast=None):
    if not isinstance(resname, VectorAccess):
        resname = VectorAccess(resname)
    if not isinstance(varname, VectorAccess):
        varname = VectorAccess(varname)
    if not isinstance(mult, MatrixAccess):
        mult = MatrixAccess(mult)
    if not isinstance(add, (VectorAccess,type(None))):
        add = VectorAccess(add)
    if mult_data.shape[0] == 0:
        if add is None or add_data.shape[0] == 0:
            return ""
        else:
            ns.add(MatrixDefinition(add, add_data))
            e = "%s" % add
    else:
        if not matrix_is_identity(mult_data):
            ns.add(MatrixDefinition(mult, mult_data))
            e = "%s * %s" % (mult, varname)
        else:
            e = "%s" % varname
        if add is not None and not matrix_is_zero(add_data):
            ns.add(MatrixDefinition(add, add_data))
            e += " + %s" % add
    if cast is not None:
        e = "(%s).cast<%s>()" % (e, cast)
    return "%s = %s;" % (resname, e)



#
# python script to generate icons form base icons
# in the first place this can be seen as documentation
#
from gtk.gdk import (pixbuf_new_from_file, Pixbuf, INTERP_HYPER,
                     PIXBUF_ROTATE_CLOCKWISE)
pb = pixbuf_new_from_file

def sliders_knobs():
    pb("hslider.png").scale_simple(
        40, 6, INTERP_HYPER).save("minislider.png","png")

    pb("hslider.png").rotate_simple(
        PIXBUF_ROTATE_CLOCKWISE).flip(True).save("vslider.png","png")

    # shorten the original eqslider and store under the same name:
    #b = pb("eqslider.png")
    #b.subpixbuf(0,8,b.get_width(),105).save("eqslider.png","png")

    for i in range(6):
        f = "knob%d.png" % (i+1)
        pb(f).scale_simple(25,25,INTERP_HYPER).save("small"+f,"png")


def transform_pixbuf(p, dest_size, scale):
    dest = Pixbuf(p.props.colorspace, p.props.has_alpha,
                  p.props.bits_per_sample, dest_size, dest_size)
    dest.fill(0)
    s = dest_size * scale / max(p.get_width(), p.get_height())
    offset_x = int(round((dest_size - p.get_width() * s) / 2))
    offset_y = int(round((dest_size - p.get_height() * s) / 2))
    dest_width = int(round(min(dest_size-offset_x, s*p.get_width())))
    dest_height = int(round(min(dest_size-offset_y, s*p.get_height())))
    p.scale(dest, offset_x, offset_y, dest_width, dest_height, offset_x, offset_y, s, s, INTERP_HYPER)
    return dest

def make_glade_icon(f, name=None, offset_x=0, offset_y=0,
                    width=None, height=None, scale=1.0):
    p = pb(f)
    if not width:
        width = p.get_width()
    if not height:
        height = p.get_height()
    if offset_x != 0 or offset_y != 0 or width != p.get_width() or height != p.get_height():
        p = p.subpixbuf(offset_x, offset_y, width, height)
    if name is None:
        name = f
    else:
        name = name + ".png"
    pre = "../glade-gxw/glade-icons/hicolor/%s/actions/widget-gxwidgets-"
    ps = transform_pixbuf(p, 16, scale)
    ps.save(pre%"16x16"+name,"png")
    ps = transform_pixbuf(p, 22, scale)
    ps.save(pre%"22x22"+name,"png")

def glade_icons():
    make_glade_icon("gxknob.png","gxbigknob")
    make_glade_icon("gxknob.png","gxsmallknob",scale=0.7)
    make_glade_icon("gxhslider.png")
    make_glade_icon("gxvslider.png")
    make_glade_icon("gxswitch.png")
    make_glade_icon("gxeqslider.png", scale=0.9)
    make_glade_icon("gxminislider.png", scale=0.8)
    make_glade_icon("gxtoggleimage.png")
    make_glade_icon("gxwheel.png")
    make_glade_icon("gxpaintbox.png")
    make_glade_icon("gxradiobutton.png")
    make_glade_icon("gxfastmeter.png")
    make_glade_icon("gxselector.png")
    make_glade_icon("gxwaveview.png")
    make_glade_icon("gxtuner.png")
    make_glade_icon("gxiredit.png")
    make_glade_icon("gxvaluedisplay.png")
    make_glade_icon("gxmeterscale.png")

if __name__ == "__main__":
    sliders_knobs()
    glade_icons()



import os, re, math, pickle
from pylab import *
from scipy.signal import freqz, lfilter
from scipy import optimize

def display_parameter(filt):
    print "Parameter:"
    print "\n".join(["    %s=%s [%s..%s]" % ((k,filt[k])+filt.get_range(k))
                   for k in filt.keys()])

def freqz_scaled(v, rate, rng, n = 2400):
    w, h = freqz(v, worN=n)
    scale = lambda freq: int(round(2 * n * freq / rate))
    n0 = scale(rng[0])
    n1 = scale(rng[1])
    w = w[n0:n1+1]
    h = h[n0:n1+1]
    w *= rate / (2*pi)
    return w, h

def parabolaTurningPoint(w, y, h=None):
    idx = y.argmax()
    yTop = y[idx-1] - y[idx+1]
    yBottom = y[idx+1] + y[idx-1] - 2 * y[idx];
    if yBottom != 0.0:
        x0 = yTop / (2 * yBottom)
    else:
        x0 = 0
    if h is not None:
        if x0 > 0:
            hh = h[idx] + x0 * (h[idx+1]-h[idx])
        else:
            hh = h[idx-1] + x0 * (h[idx]-h[idx-1])
        return w[idx] + x0*(w[idx+1]-w[idx]), hh
    else:
        return w[idx] + x0*(w[idx+1]-w[idx]), y[idx] - (yBottom / 2) * x0 * x0

def db_point(w, h, db):
    d = h - db
    return parabolaTurningPoint(w, -d * d, h)

def plotone(filt, freq, para, impulse, clr='--', fs=44100):
    filt[para] = freq
    filt.compute(zeros(10000,dtype=float32))
    w, h = freqz_scaled(filt.compute(impulse), fs, [30, 15000], 15000)
    h = 20*log10(abs(h))
    wmax, hmax = parabolaTurningPoint(w, h)
    wmax3, hmax3 = db_point(w, h, hmax-3)
    q = wmax/wmax3
    Q = 1/abs(q-1/q)
    print "%.2f %.2f %g %g" % (freq, wmax, hmax, Q)
    semilogx(w, h, clr, label="%.3f" % freq)
    return (freq, wmax, hmax, Q)

def plot_figure(filt, name, para, rg, clr="-", fs=44100):
    display_parameter(filt)
    impulse = zeros(10000,dtype=float32)
    impulse[0] = 1
    dflt = filt[para]
    #subplot(2,2,1)
    title('Guitarix crybaby2')
    ylabel("gain (dB)")
    xlabel("Frequency (Hz)")
    sn = False
    res = []
    for freq in rg:
        res.append(plotone(filt, freq, para, impulse, clr, fs))
    #legend(title="Param.: %s" % name, fancybox=True, loc='upper right')
    #a = array(res);
    #subplot(2,2,2)
    #plot(1/(1.06-a[:,0]), a[:,1])
    #subplot(2,2,3)
    #plot(a[:,0], a[:,2])
    #subplot(2,2,4)
    #plot(a[:,0], a[:,3])
    #show()

def plot_figure_libcrybaby(clr="-",fs=44100):
    from pluginloader import Plugin
    filt = Plugin("../build/default/src/plugins/libcrybaby.so")
    filt['crybaby2.refvolt'] = 0.1
    para = 'crybaby2.hotpotz'
    name = filt.get_var_attr(para)[0]
    rg = log10(linspace(*np.power(10,filt.get_range(para)), num=10))
    filt.init(fs)
    plot_figure(filt, name, para, rg, clr, fs)

def plot_figure_dunwah(clr="-", fs=44100):
    import dunwah
    filt = dunwah.dsp()
    para = 'wah'
    name = "DunWah"
    #rg = linspace(*filt.get_range(para), num=10)
    rg = log10(linspace(*np.power(10,filt.get_range(para)), num=10))
    filt.init(fs)
    plot_figure(filt, name, para, rg, clr, fs)

def plot_figure_crybaby(clr="-", fs=44100):
    import crybaby
    para = '.crybaby.wah'
    filt = crybaby.dsp()
    name = "Pot"
    #rg = linspace(*filt.get_range(para), num=10)
    rg = log10(linspace(*np.power(10,filt.get_range(para)), num=10))
    filt.init(fs)
    plot_figure(filt, name, para, rg, clr, fs)

def plot_figure_autowah(clr="-", fs=44100):
    import ladspaloader
    filt = ladspaloader.Ladspa("/usr/lib/ladspa/autowah.so", fs)
    filt["Drive"] = 0
    filt["Decay"] = 0
    filt["Range"] = 0
    filt["Mix"] = 1
    name = "Freq"
    para = "Freq"
    rg = linspace(0, 1, num=10)
    plot_figure(filt, name, para, rg, clr, fs)

def save_octave(name, var, fd):
    assert isinstance(var, ndarray) and len(var.shape) == 1
    print >>fd, "# name:", name
    iscompl = iscomplexobj(var)
    if iscompl:
        print >>fd, "# type: complex matrix"
    else:
        print >>fd, "# type: matrix"
    print >>fd, "# rows: 1"
    print >>fd, "# columns:", var.shape[0]
    if iscompl:
        for v in var:
            print >>fd, (v.real, v.imag),
    else:
        for v in var:
            print >>fd, v,
    print >>fd

def load_octave(fd):
    var = []
    d = {}
    for l in fd:
        if l.startswith("#"):
            m = re.match("# *([a-z]+): *(.*)\n", l)
            if not m:
                continue
            d[m.group(1)] = m.group(2)
        else:
            var.append(array([float(v) for v in l.split()]))
    return var

def invfreqz(w, h):
    fd = file("/tmp/out", "w");
    save_octave("F", w, fd)
    save_octave("H", h, fd)
    fd.close()
    p = os.popen("octave -q","w")
    p.write(
        "load /tmp/out;\n"
        "wt = 1 ./ F.^2;\n"
        "[B,A] = invfreqz(H,F,3,3,wt);\n"
        "save /tmp/out B A;\n")
    p.close()
    fd = file("/tmp/out");
    res = load_octave(fd)
    fd.close()
    os.remove("/tmp/out")
    return res

def run_filter(filt,impulse,fs):
    # skip transients
    filt.compute(zeros(10000,dtype=float32))
    # calculate frequency response
    w, h = freqz(filt.compute(impulse), worN=15000)
    # calculate filter coefficients
    k1 = int(round(len(h)*10.0/fs))
    k2 = int(round(len(h)*20000.0/fs))
    B, A = invfreqz(w[k1:k2+1], h[k1:k2+1])
    return B, A, h

def estimate(rg, filt, para, impulse, fs):
    aa = zeros(len(rg))
    Q = zeros(len(rg))
    maxh = zeros(len(rg))
    real_roots = zeros(len(rg))
    for i, freq in enumerate(rg):
        filt[para] = freq
        B, A, h = run_filter(filt, impulse, fs)
        maxh[i] = max(abs(h))
        # use a constant B
        #print B/B[0]
        if i == (len(rg)+1)/2:
            Bconst = B
            freq_const = freq
        # use just the two conjugate roots for the biquad
        # the real root is nearly constant
        bq_roots = []
        for j, v in enumerate(roots(A)):
            if v.imag == 0:
                real_roots[i] = v.real
            else:
                bq_roots.append(v)
        A_bq = poly(bq_roots)
        # calculate pole frequency and quality factor
        R = sqrt(A_bq[2])
        theta = math.acos(A_bq[1]/(-2*R))
        aa[i] = fs * theta
        Q[i] = theta / (2 - 2*R)

        print 'Pole frequency = %f Hz' % (aa[i]/(2*pi))
        #print 'Q = %f' % Q[i]
        #print "R =", R

    return aa, Q, maxh, sum(real_roots) / len(real_roots), Bconst, freq_const

def fit_param(rg, aa, Q, fs):
    #order = 3
    order = 6
    def f(t):
        xx = 1/(t-aa)
        a = polyfit(rg, xx, order)
        def ff(x, *p):
            return t - 1/polyval(p, x)
        a = optimize.curve_fit(ff, rg, aa, a)
        a = a[0]
        res = sum((xx-polyval(a,rg))**2)/sum(xx*xx)**2
        return res
    off = optimize.fmin(f, 0.04*fs, disp=False)[0]
    f = 1/(off-aa)
    a = polyfit(rg, f, order)
    def ff(x, *p):
        return off - 1/polyval(p, x)
    a = optimize.curve_fit(ff, rg, aa, a)
    a = a[0]
    a1A = a
    #order = 4
    order = 5
    a = polyfit(rg, Q, order)
    qA = a
    return off, a1A, qA

def show_param(rg, off, g_off, aa, a1A, Q, qA, gain, gA):
    subplot(2,2,1)
    f = 1/(off-aa)
    plot(rg, f,"b")
    plot(rg,polyval(a1A,rg),"r")
    subplot(2,2,2)
    plot(rg, Q,"g")
    plot(rg, polyval(qA,rg), "r")
    subplot(2,2,3)
    plot(rg, 1/(g_off-gain), "b")
    plot(rg, polyval(gA,rg), "r")
    show()

def string_polyval(a, var):
    s = ""
    for i, v in enumerate(a):
        if i == 0:
            s = "%g" % v
        elif i == 1:
            s = "%s*%s+%g" % (s, var, v)
        else:
            s = "(%s)*%s+%g" % (s, var, v)
    return s

def make_gain(rg, off, a1A, qA, maxh, z1, Bconst, fs):
    gain = zeros(len(rg))
    for i, freq in enumerate(rg):
        q  = polyval(qA, freq)
        a1 = (off - 1 / polyval(a1A, freq)) / fs
        r = 1 - a1/(2*q)
        A = array((1, -2*r*math.cos(a1), r*r))
        rr = list(roots(A))
        rr.append(z1)
        A = poly(rr)
        w1, h1 = freqz(Bconst, A, worN=15000)
        gain[i] = maxh[i]/max(abs(h1))
    #order = 1
    order = 5
    def ff(t):
        xx = 1/(t-gain)
        a = polyfit(rg, xx, order)
        def f(x, *p):
            return t-1/polyval(p, x)
        a = optimize.curve_fit(f, rg, gain, a)
        a = a[0]
        return sum((xx-polyval(a,rg))**2)
    g_off = optimize.fmin(ff, -0.007, disp = False)
    def f(x, *p):
        return g_off-1/polyval(p, x)
    a = polyfit(rg, 1/(g_off-gain), order)
    a = optimize.curve_fit(f, rg, gain, a)
    a = a[0]
    return gain, g_off, a

def estimate_SR(filt, para, freq_const, impulse):
    lz1 = []
    lz2 = []
    lz3 = []
    lg = []
    ss = linspace(44100, 96000, 10)
    for fs in ss:
        filt.init(fs)
        filt[para] = freq_const
        B, A, h = run_filter(filt, impulse, fs)
        rr = roots(B)
        lz1.append(rr[0])
        lz2.append(rr[1])
        lz3.append(rr[2])
        lg.append(B[0])
    z1 = sum(lz1)/len(lz1)
    z2A = polyfit(ss, lz2, 2)
    z3A = polyfit(ss, lz3, 2)
    gc = polyfit(ss, lg, 2)
    return z1, z2A, z3A, gc

def output_freqz_libcrybaby():
    from pluginloader import Plugin
    filt = Plugin("../build/default/src/plugins/libcrybaby.so")
    filt['crybaby2.refvolt'] = 0.1
    para = 'crybaby2.hotpotz'
    name = filt.get_var_attr(para)[0]
    fs = 44100
    filt.init(fs)
    impulse = zeros(10000,dtype=float32)
    impulse[0] = 1
    rg = log10(linspace(*np.power(10,filt.get_range(para)), num=20))
    if False:
        aa, Q, maxh, p1, Bconst, freq_const = estimate(rg, filt, para, impulse, fs)
        pickle.dump((aa,Q,maxh,p1,Bconst,freq_const), file("p.out","w"))
    else:
        aa, Q, maxh, p1, Bconst, freq_const = pickle.load(file("p.out"))
    z1, z2A, z3A, gcA = estimate_SR(filt, para, freq_const, impulse)
    off, a1A, qA = fit_param(rg, aa, Q, fs)
    gain, g_off, gA = make_gain(rg, off, a1A, qA, maxh, p1, Bconst, fs)
    #show_param(rg, off, g_off, aa, a1A, Q, qA, gain, gA)
    rg = log10(linspace(*np.power(10,filt.get_range(para)), num=10))
    filt.init(fs)
    for i, freq in enumerate(rg):
        filt[para] = freq
        filt.compute(zeros(10000,dtype=float32))
        w, h = freqz(filt.compute(impulse), worN=15000)

        q  = polyval(qA, freq)
        a1 = (off - 1 / polyval(a1A, freq)) / fs
        g = g_off - 1 / polyval(gA, freq)
        gc = polyval(gcA, fs)

        r = 1 - a1/(2*q)
        A = array((1, -2*r*cos(a1), r*r))
        A = poly(list(roots(A))+[p1])
        B = poly([polyval(z2A,fs),polyval(z3A,fs),z1])
        w1, h1 = freqz(B*g*gc, A, worN=15000)

        semilogx((w*fs/(2*pi))[1:],  20*log10(abs(h[1:])), "b")
        semilogx((w1*fs/(2*pi))[1:], 20*log10(abs(h1[1:])), "r")

    print "theta2pi = (%g - 1000 / (%s)) / SR;" % (off, string_polyval(a1A*1000, "wah"))
    print "Q = %s;" % string_polyval(qA, "wah")
    print "g = %g - 1 / (%s);" % (g_off, string_polyval(gA, "wah"))
    print "gc = %s;" % string_polyval(gcA, "SR")
    print "p1 = exp(-1000/(%g*SR));" % (-1000/(fs*log(p1)))
    print "z1 = %g;" % z1
    print "z2 = %s;" % string_polyval(z2A, "SR")
    print "z3 = %s;" % string_polyval(z3A, "SR")

    show()

def compare_dynamics(freq, fs):
    global title
    title("Freq = %g" % freq)

    lfo_freq = 5.0
    block_size = 64
    time = 1
    block_cnt = (time*fs+block_size-1)/block_size
    samples = block_cnt * block_size

    def sine(freq, div=1):
        return array(sin(2*pi*linspace(0,freq*time,samples / div)),dtype=float32)

    tm = linspace(0,samples/float(fs),samples)
    lfo = 0.5 + sine(lfo_freq,block_size) * 0.5
    signal = sine(freq).reshape(block_cnt,block_size)
    tt = linspace(0,1,samples / 100)

    def sweep(filt, para, clr):
        filt[para] = 0.0
        filt.compute(zeros(10000,float32))
        out = zeros((block_cnt,block_size))
        for i in range(block_cnt):
            filt[para] = lfo[i]
            out[i] = filt.compute(signal[i])
        t = 0.999
        ampl = lfilter([1-t],[1,-t],abs(out.reshape(samples)))
        plot(tm[4000:], 20*log10(ampl[4000:]), clr)

    import dunwah
    filt = dunwah.dsp()
    filt.init(fs)
    para = "wah"

    sweep(filt, para, "r")

    from pluginloader import Plugin
    filt = Plugin("../build/default/src/plugins/libcrybaby.so")
    filt.init(fs)
    filt['crybaby2.refvolt'] = 0.1
    para = 'crybaby2.hotpotz'

    sweep(filt, para, "b")

def dynamics_figure(fs):
    subplot(221)
    compare_dynamics(410, fs)
    subplot(222)
    compare_dynamics(500, fs)
    subplot(223)
    compare_dynamics(800, fs)
    subplot(224)
    compare_dynamics(1000, fs)

def compare_freqz(fs):
    plot_figure_dunwah("r", fs)
    plot_figure_libcrybaby("b", fs)

if __name__ == "__main__":
    fs = 44100
    #fs = 48000*2
    try:
        compare_freqz(fs)
        #output_freqz_libcrybaby()
        #dynamics_figure(fs)
        show()
    except KeyboardInterrupt:
        pass



#! /usr/bin/env python
# encoding: utf-8

import os,sys
import Task,Utils
from Configure import conftest
ar_str='${AR} ${ARFLAGS} ${AR_TGT_F}${TGT} ${AR_SRC_F}${SRC}'
cls=Task.simple_task_type('static_link',ar_str,color='YELLOW',ext_in='.o',ext_out='.bin',shell=False)
cls.maxjobs=1
cls.install=Utils.nada
old=cls.run
def wrap(self):
	try:os.remove(self.outputs[0].abspath(self.env))
	except OSError:pass
	return old(self)
setattr(cls,'run',wrap)
def detect(conf):
	conf.find_program('ar',var='AR')
	conf.find_program('ranlib',var='RANLIB')
	conf.env.ARFLAGS='rcs'
def find_ar(conf):
	v=conf.env
	conf.check_tool('ar')
	if not v['AR']:conf.fatal('ar is required for static libraries - not found')

conftest(find_ar)



#! /usr/bin/env python
# encoding: utf-8

import TaskGen,Utils,Task,Options
from Logs import error
from TaskGen import before,after,taskgen,feature
flag_vars=['FLAGS','ASSEMBLIES']
def init_cs(self):
	Utils.def_attrs(self,flags='',assemblies='',resources='',uselib='')
def apply_uselib_cs(self):
	if not self.uselib:
		return
	global flag_vars
	for var in self.to_list(self.uselib):
		for v in self.flag_vars:
			val=self.env[v+'_'+var]
			if val:self.env.append_value(v,val)
def apply_cs(self):
	try:self.meths.remove('apply_core')
	except ValueError:pass
	for i in self.to_list(self.assemblies)+self.env['ASSEMBLIES']:
		self.env.append_unique('_ASSEMBLIES','/r:'+i)
	for i in self.to_list(self.resources):
		self.env.append_unique('_RESOURCES','/resource:'+i)
	self.env['_TYPE']=getattr(self,'type','exe')
	self.env.append_unique('_FLAGS',self.to_list(self.flags))
	self.env.append_unique('_FLAGS',self.env.FLAGS)
	nodes=[self.path.find_resource(i)for i in self.to_list(self.source)]
	self.create_task('mcs',nodes,self.path.find_or_declare(self.target))
Task.simple_task_type('mcs','${MCS} ${SRC} /target:${_TYPE} /out:${TGT} ${_FLAGS} ${_ASSEMBLIES} ${_RESOURCES}',color='YELLOW')
def detect(conf):
	csc=getattr(Options.options,'cscbinary',None)
	if csc:
		conf.env.MCS=csc
	conf.find_program(['gmcs','mcs'],var='MCS')
def set_options(opt):
	opt.add_option('--with-csc-binary',type='string',dest='cscbinary')

feature('cs')(init_cs)
feature('cs')(apply_uselib_cs)
after('init_cs')(apply_uselib_cs)
feature('cs')(apply_cs)
after('apply_uselib_cs')(apply_cs)
before('apply_core')(apply_cs)



#! /usr/bin/env python
# encoding: utf-8

import os
import Task,Options,Utils
from TaskGen import before,feature,after
from Configure import conf
def init_rubyext(self):
	self.default_install_path='${ARCHDIR_RUBY}'
	self.uselib=self.to_list(getattr(self,'uselib',''))
	if not'RUBY'in self.uselib:
		self.uselib.append('RUBY')
	if not'RUBYEXT'in self.uselib:
		self.uselib.append('RUBYEXT')
def apply_ruby_so_name(self):
	self.env['shlib_PATTERN']=self.env['rubyext_PATTERN']
def check_ruby_version(conf,minver=()):
	if Options.options.rubybinary:
		conf.env.RUBY=Options.options.rubybinary
	else:
		conf.find_program("ruby",var="RUBY",mandatory=True)
	ruby=conf.env.RUBY
	try:
		version=Utils.cmd_output([ruby,'-e','puts defined?(VERSION) ? VERSION : RUBY_VERSION']).strip()
	except:
		conf.fatal('could not determine ruby version')
	conf.env.RUBY_VERSION=version
	try:
		ver=tuple(map(int,version.split(".")))
	except:
		conf.fatal('unsupported ruby version %r'%version)
	cver=''
	if minver:
		if ver<minver:
			conf.fatal('ruby is too old')
		cver=".".join([str(x)for x in minver])
	conf.check_message('ruby',cver,True,version)
def check_ruby_ext_devel(conf):
	if not conf.env.RUBY:
		conf.fatal('ruby detection is required first')
	if not conf.env.CC_NAME and not conf.env.CXX_NAME:
		conf.fatal('load a c/c++ compiler first')
	version=tuple(map(int,conf.env.RUBY_VERSION.split(".")))
	def read_out(cmd):
		return Utils.to_list(Utils.cmd_output([conf.env.RUBY,'-rrbconfig','-e',cmd]))
	def read_config(key):
		return read_out('puts Config::CONFIG[%r]'%key)
	ruby=conf.env['RUBY']
	archdir=read_config('archdir')
	cpppath=archdir
	if version>=(1,9,0):
		ruby_hdrdir=read_config('rubyhdrdir')
		cpppath+=ruby_hdrdir
		cpppath+=[os.path.join(ruby_hdrdir[0],read_config('arch')[0])]
	conf.check(header_name='ruby.h',includes=cpppath,mandatory=True,errmsg='could not find ruby header file')
	conf.env.LIBPATH_RUBYEXT=read_config('libdir')
	conf.env.LIBPATH_RUBYEXT+=archdir
	conf.env.CPPPATH_RUBYEXT=cpppath
	conf.env.CCFLAGS_RUBYEXT=read_config("CCDLFLAGS")
	conf.env.rubyext_PATTERN='%s.'+read_config('DLEXT')[0]
	flags=read_config('LDSHARED')
	while flags and flags[0][0]!='-':
		flags=flags[1:]
	if len(flags)>1 and flags[1]=="ppc":
		flags=flags[2:]
	conf.env.LINKFLAGS_RUBYEXT=flags
	conf.env.LINKFLAGS_RUBYEXT+=read_config("LIBS")
	conf.env.LINKFLAGS_RUBYEXT+=read_config("LIBRUBYARG_SHARED")
	if Options.options.rubyarchdir:
		conf.env.ARCHDIR_RUBY=Options.options.rubyarchdir
	else:
		conf.env.ARCHDIR_RUBY=read_config('sitearchdir')[0]
	if Options.options.rubylibdir:
		conf.env.LIBDIR_RUBY=Options.options.rubylibdir
	else:
		conf.env.LIBDIR_RUBY=read_config('sitelibdir')[0]
def set_options(opt):
	opt.add_option('--with-ruby-archdir',type='string',dest='rubyarchdir',help='Specify directory where to install arch specific files')
	opt.add_option('--with-ruby-libdir',type='string',dest='rubylibdir',help='Specify alternate ruby library path')
	opt.add_option('--with-ruby-binary',type='string',dest='rubybinary',help='Specify alternate ruby binary')

feature('rubyext')(init_rubyext)
before('apply_incpaths','apply_type_vars','apply_lib_vars','apply_bundle')(init_rubyext)
after('default_cc','vars_target_cshlib')(init_rubyext)
feature('rubyext')(apply_ruby_so_name)
before('apply_link')(apply_ruby_so_name)
conf(check_ruby_version)
conf(check_ruby_ext_devel)



#! /usr/bin/env python
# encoding: utf-8

import os,sys,imp,types,ccroot
import optparse
import Utils,Configure,Options
from Logs import debug
c_compiler={'win32':['msvc','gcc'],'cygwin':['gcc'],'darwin':['gcc'],'aix':['xlc','gcc'],'linux':['gcc','icc','suncc'],'sunos':['gcc','suncc'],'irix':['gcc'],'hpux':['gcc'],'gnu':['gcc'],'default':['gcc']}
def __list_possible_compiler(platform):
	try:
		return c_compiler[platform]
	except KeyError:
		return c_compiler["default"]
def detect(conf):
	try:test_for_compiler=Options.options.check_c_compiler
	except AttributeError:conf.fatal("Add set_options(opt): opt.tool_options('compiler_cc')")
	orig=conf.env
	for compiler in test_for_compiler.split():
		conf.env=orig.copy()
		try:
			conf.check_tool(compiler)
		except Configure.ConfigurationError as e:
			debug('compiler_cc: %r'%e)
		else:
			if conf.env['CC']:
				orig.table=conf.env.get_merged_dict()
				conf.env=orig
				conf.check_message(compiler,'',True)
				conf.env['COMPILER_CC']=compiler
				break
			conf.check_message(compiler,'',False)
			break
	else:
		conf.fatal('could not configure a c compiler!')
def set_options(opt):
	build_platform=Utils.unversioned_sys_platform()
	possible_compiler_list=__list_possible_compiler(build_platform)
	test_for_compiler=' '.join(possible_compiler_list)
	cc_compiler_opts=opt.add_option_group("C Compiler Options")
	cc_compiler_opts.add_option('--check-c-compiler',default="%s"%test_for_compiler,help='On this platform (%s) the following C-Compiler will be checked by default: "%s"'%(build_platform,test_for_compiler),dest="check_c_compiler")
	for c_compiler in test_for_compiler.split():
		opt.tool_options('%s'%c_compiler,option_group=cc_compiler_opts)




# -*- coding: utf-8 -*-
from __future__ import division

for pyname, package in (
        ('scipy', 'python-scipy'),
        ('sympy', 'python-sympy'),
        ('pylab', 'python-matplotlib'),
        ):
    try:
        __import__(pyname)
    except ImportError:
        raise SystemExit("Need module %s (Debian package: %s)" % (pyname, package))

import sys, itertools, fractions, os, argparse, math, logging
from cStringIO import StringIO
from scipy.interpolate import LSQUnivariateSpline
try:
    from scipy.interpolate import PchipInterpolator
except ImportError:
    print "pchip interpolation not available (scipy version too old)"
import pylab as pl
import numpy as np
from scipy.signal import correlate
from scipy.interpolate import splev
import dk_simulator, dk_lib, circ, models


class KnotData(object):

    def __init__(self, knots, mapping, sl, order):
        self.knots = knots
        self.mapping = mapping
        self.slice_data = sl
        if order is None:
            self.tp = None
            self.order = None
            self.bbox = None
        else:
            if isinstance(order, int):
                self.tp = 's'
                self.order = order
            else:
                self.tp = order[0]
                self.order = order[1]
            self.bbox = [self.knots[0], self.knots[-1]]

    def cut_mapping(self):
        m = self.mapping
        i0, i1 = np.searchsorted(m, (m[0]+1, m[-1]))
        i0 -= 1
        i1 += 1
        self.bbox[0] += i0 / self.inv_h()
        self.bbox[1] -= (len(m) - i1) / self.inv_h()
        self.mapping = m[i0:i1]

    def used(self):
        return self.order is not None

    def inv_h(self):
        sl = self.slice_data
        return (sl.step.imag - 1) / (sl.stop - sl.start)

    def count(self):
        return len(self.mapping) + 1

    def get_knots(self):
        k = self.order
        return self.knots[k:-k]

    def get_knot_grid(self):
        d = self.slice_data
        k = self.order
        if k is None:
            return np.array(((d.start+d.stop)*0.5,))
        else:
            a = self.knots[k//2:-k//2].copy()
            n = d.step.imag
            if n == 1:
                n = k
            h = (d.stop - d.start) / (n - 1)
            for i in range(k//2):
                a[i] = d.start + i * h
                a[-(i+1)] = d.stop - i * h
            return a

    def num_coeffs(self):
        return len(self.knots) - self.order

    def get_bbox(self):
        return self.bbox

    def get_order(self):
        return self.order

class TensorSpline(object):

    def __init__(self, func, ranges, basegrid):
        self.func = func
        self.ranges = ranges
        self.basegrid = basegrid
        self.knot_data = np.empty((len(basegrid), len(ranges)), dtype=object)
        self.get_grid_estimate()
        self.find_knots()
        self.save_grid_estimate()

    @staticmethod
    def max_idx_to_maptype(max_idx):
        if max_idx >= 2**16:
            return "int"
        elif max_idx >= 2**8:
            return "unsigned short"
        else:
            return "unsigned char"

    def grid_estimate_fname(self):
        return "%s_grid.cache" % self.func.comp_id

    def get_grid_estimate(self):
        try:
            self.grid_estimate = pickle.load(open(self.grid_estimate_fname()))
        except (IOError, EOFError):
            self.grid_estimate = np.empty((len(self.basegrid), len(self.ranges)), dtype=object)
            for v, g in enumerate(self.basegrid):
                self.grid_estimate[v][:] = [np.mgrid[slice(r[0], r[1], 0.5j * t[0])] for r, t in zip(self.ranges,g[0])]

    def save_grid_estimate(self):
        with open(self.grid_estimate_fname(),"w") as f:
            pickle.dump(self.grid_estimate, f)

    @staticmethod
    def insert_points(a, n, kd):
        o = []
        for x, k in zip(a, kd):
            if k.tp == 'h':
                o.append(x)
                continue
            i = len(x)
            b = np.zeros(i + n * (i -  1))
            b[::n+1] = x
            for j in range(n):
                w = (j+1) / (n+1)
                b[j+1::n+1] = x[:-1] * w + x[1:] * (1-w)
            o.append(b)
        return o

    def calc_coeffs(self):
        coeff = []
        for v, (g, kdata) in enumerate(zip(self.basegrid, self.knot_data)):
            coords = []
            for kd in kdata:
                if kd.used():
                    if kd.tp in ('s', 'pp'):
                        k = kd.get_order()
                        x = kd.knots.copy()
                        x[:k] = np.linspace(x[0], x[k], k, False)
                        x[-k:] = np.linspace(x[-(k+1)], x[-1], k, False)
                    else:
                        assert kd.tp == 'h'
                        k = kd.get_order() - 1
                        x = kd.knots[k:-k]
                else:
                    x = kd.get_knot_grid()
                coords.append(x)
            add_num_points = 0
            coords = self.insert_points(coords, add_num_points, self.knot_data[v])
            fnc = self.calc_grid(self.make_grid(coords), g[1], g[2])[v]
            for n in range(len(self.ranges)):
                kd = self.knot_data[v,n]
                if not kd.used():
                    continue
                s = list(fnc.shape)
                if kd.tp == 'h':
                    s[n] = 4 * (len(coords[n]) - 1)
                else:
                    s[n] = kd.num_coeffs()
                out = np.empty(s)
                s[n] = 1
                x = coords[n]
                kn = kd.get_knots()
                bb = kd.get_bbox()
                k = kd.get_order() - 1
                for i in np.ndindex(tuple(s)):
                    i = list(i)
                    i[n] = slice(None)
                    if kd.tp in ('s', 'pp'):
                        ss = LSQUnivariateSpline(x, fnc[i], kn, bbox=bb, k=k)
                        out[i] = ss.get_coeffs()
                    else:
                        ss = PchipInterpolator(x, fnc[i])
                        def mklist(kr):
                            c = kr.c[:,0]
                            return [c[3], c[2]-c[3]*(kr.xi[-1]-kr.xi[0]), c[1], c[0]]
                        out[i] = np.array([mklist(kr) for kr in ss.polynomials]).flat
                fnc = out
            coeff.append(fnc)
        return coeff

    @staticmethod
    def get_interpolating_knots(x, k):
        x_e = x[-1]
        x_e += (x_e - x[-2]) * 0.1
        start = k // 2
        end = -(k - start)
        knots = np.zeros((3,len(x)-k))
        knots[0] = x[start:end]
        knots[1] = np.arange(start, len(x)+end)
        if k % 2:
            knots[0,:-1] += knots[0,1:]
            knots[0,-1] += x[end]
            knots[0] *= 0.5
        return knots, (x[0], x_e)

    @staticmethod
    def spline_diff(n, x, kn, bbox, k, fnc):
        s = list(fnc.shape)
        s[n] = 1
        out = np.empty_like(fnc)
        for idx in np.ndindex(tuple(s)):
            idx = list(idx)
            idx[n] = slice(None)
            f = fnc[idx]
            ss = LSQUnivariateSpline(x, f, kn, bbox=bbox, k=k-1)
            out[idx] = abs(ss(x) - f)
        return out

    @staticmethod
    def max_df(n, x, kn, bbox, k, fnc):
        s = list(fnc.shape)
        s[n] = 1
        mx = 0
        for idx in np.ndindex(tuple(s)):
            idx = list(idx)
            idx[n] = slice(None)
            f = fnc[idx]
            ss = LSQUnivariateSpline(x, f, kn, bbox=bbox, k=k-1)
            mx = max(mx, np.amax(abs(ss(x) - f)))
        return mx #/ np.ptp(fnc)

    def make_grid(self, coords, sparse=False):
        if len(coords) == 1:
            r = coords[0]
            res = np.empty((1, r.shape[0]))
            res[0] = r
        else:
            r = np.meshgrid(*coords, indexing="ij", sparse=sparse)
            res = np.empty((len(coords),)+r[0].shape)
            res[:] = r
        return res

    def calc_grid(self, grd, pre, post):
        if post:
            pre_grd = grd.copy()
        if pre:
            grd = pre(grd)
        grd_shape = grd.shape
        numpoints = np.product(grd_shape[1:])
        grd = grd.reshape(grd_shape[0], numpoints)
        fnc = self.func(grd.T, True).T
        fnc = fnc.reshape(fnc.shape[:1]+tuple(grd_shape[1:]))
        if post:
            fnc = post(pre_grd, fnc)
        return fnc

    @staticmethod
    def max_diff(n, kn, pos, val):
        ki = np.unique(np.digitize(pos, kn), True)[1]
        idx = [slice(None)]*val.ndim
        l = []
        for i, j in itertools.izip_longest(ki, ki[1:]):
            idx[n] = slice(i, j)
            l.append(np.amax(val[idx]))
        a = np.array(l)
        return a[1:] + a[:-1]

    @staticmethod
    def mk_result(kn, bbox, k, r):
        idx = np.zeros(len(kn[1])+2, dtype=np.int32)
        idx[1:-1] = kn[1]
        idx[-1] = r.step.imag
        f = reduce(fractions.gcd, idx)
        idx /= f
        a = np.empty(idx[-1], dtype=np.int32)
        for m, (i, j) in enumerate(itertools.izip_longest(idx[:-1], idx[1:])):
            a[i:j] = m
        a += k-1
        return KnotData(np.pad(kn[0], k, 'constant', constant_values=bbox), a, slice(r.start, r.stop, r.step/f), k)

    def get_support(self, i, fnc, dlim):
        def apply_subcube(op, a, axis):
            j = a.shape[axis]
            return op(np.rollaxis(a, axis, 1).reshape(j, a.size/j), axis=1)
        dfnc = np.diff(fnc, 1, axis=i)
        supp1 = apply_subcube(np.amax, abs(dfnc), i) >= dlim
        nz1 = np.flatnonzero(supp1)
        d2fnc = np.diff(dfnc, 1, axis=i)
        d2lim = 1e-7
        supp2 = apply_subcube(np.amax, abs(d2fnc), i) >= d2lim
        nz2 = np.flatnonzero(supp2)
        istart, iend, llim, rlim = None, None, None, None
        if len(nz1):
            istart = nz1[0]
            iend = nz1[-1]
        if len(nz2):
            istart = min(istart, nz2[0])
            iend = max(iend, nz2[-1])
        if istart < 2:
            istart = None
        if iend > d2fnc.shape[i]-1:
            iend = None
        if istart:
            istart -= 1
            if np.amax(apply_subcube(np.amax, abs(dfnc), i)[:istart]) < dlim:
                llim = 0
            else:
                llim = (np.amin(apply_subcube(np.amin, dfnc, i)[:istart]), np.amax(apply_subcube(np.amax, dfnc, i)[:istart]))
        if iend:
            iend += 1
            if np.amax(apply_subcube(np.amax, abs(dfnc), i)[iend:]) < dlim:
                rlim = 0
            else:
                rlim = (np.amin(apply_subcube(np.amin, dfnc, i)[:iend]), np.amax(apply_subcube(np.amax, dfnc, i)[:iend]))
                print apply_subcube(np.amax, abs(dfnc), i)[iend:]; print iend, dlim; raise SystemExit
        print "##", dlim, (istart, iend), (llim, rlim)
        return (istart, iend), (llim, rlim)

    def find_knots(self):
        for v, g in enumerate(self.basegrid):
            order = [t[1] for t in g[0] if t is not None]
            for n in range(len(self.ranges)):
                coords = self.grid_estimate[v].copy()
                coords[n] = np.mgrid[self.ranges[n][0]:self.ranges[n][1]:g[0][n][0]*2j]
                #ranges = [slice(r[0], r[1], (2j * t[0] + 1 if i == n else 0.5j * t[0])) for i, (r, t) in enumerate(zip(self.ranges,g[0]))]
                #fnc = self.calc_grid(np.mgrid[ranges], g[1], g[2])[v]
                fnc = self.calc_grid(self.make_grid(coords), g[1], g[2])[v]
                #self.get_support(n, fnc, np.amax(abs(fnc))*(ranges[n].stop-ranges[n].start)/ranges[n].step.imag * 1e-4)
                ptp = fnc.ptp()
                if np.amax(fnc.ptp(axis=n)) < 1e-6 * ptp and False: ##FIXME
                    print "%s[%d,%d]: const: %g" % (self.func.comp_id, v, n, np.amax(fnc.ptp(axis=n)) / ptp)
                    self.knot_data[v, n] = KnotData(None,None,slice(self.ranges[n][0], self.ranges[n][1], 1j),None)
                    continue
                k = order[n]
                #x = np.mgrid[ranges[n]]
                x = coords[n]
                rng = slice(self.ranges[n][0], self.ranges[n][1], g[0][n][0]*1j)
                kn, bbox = self.get_interpolating_knots(np.mgrid[rng], k)
                #e_max = 1.1 * self.max_df(n, x, kn, bbox, k, fnc)
                try:
                    e_rel = g[3]
                except IndexError:
                    e_rel = None
                if e_rel is None:
                    #e_rel = 4e-5
                    e_rel = 4e-4
                if e_rel < 0:
                    e_max = -e_rel
                else:
                    e_max = ptp * e_rel
                good = kn
                am = None
                kn0 = kn
                try:
                    opt = g[4]
                except IndexError:
                    opt = True
                while opt:
                    df = self.spline_diff(n, x, kn[0], bbox, k, fnc)
                    mx = np.amax(df)
                    if mx > e_max:
                        if am is None:
                            raise ValueError("bad approximation for %s[%d,%d]: %g > %g" % (self.func.comp_id, v, n, mx, e_max))
                        kn = good
                        kn[2,am] = 1
                    else:
                        good = kn
                    if False:
                        xd = np.square(np.diff(ss(kn[:-1], k))).argsort()+1
                    else:
                        xd = self.max_diff(n, np.pad(kn[0],(1,0),'constant',constant_values=(x[0],)), x, df).argsort()
                    for am in xd:
                        if not kn[2,am]:
                            break
                    else:
                        break
                    kn = np.append(kn[:,:am],kn[:,am+1:], axis=1)
                self.knot_data[v, n] = self.mk_result(kn, bbox, k, rng)
                self.grid_estimate[v,n] = self.knot_data[v,n].get_knot_grid()
                print "%s[%d,%d]: (%g), %d -> %d [%s]" % (self.func.comp_id, v, n, e_max, len(kn0[0]), len(kn[0]), fnc.shape)
                #print "#", fnc.shape, x.shape, e_max
                #print kn[:1]
                #pylab.plot(x, fnc[:,0,:])
                #pylab.plot(kn, np.zeros(len(kn)),"x")
                #pylab.show(); raise SystemExit


#dtp = "float"
#tiny = np.finfo(np.float32).tiny
dtp = "treal"
tiny = np.finfo(np.float64).tiny

def print_float_array(o, name, a, prefix):
    print >>o, "%s%s %s[%d] = {" % (prefix, dtp, name, a.size)
    s = "  "
    i = 0
    a = np.where(abs(a) < 100*tiny, 0, a)
    for v in a.ravel('C'):
        o.write(s+repr(float(v)))
        #o.write(s + "%.8g" % v)
        s = ","
        i += 1
        if i % 10 == 0:
            print >>o, ","
            s = "  "
    print >>o, "};"
    return a.size * np.float32().nbytes

def print_int_array(o, name, a, prefix):
    print >>o, "%sint %s[%d] = {" % (prefix, name, a.size)
    s = "  "
    i = 0
    for v in a.ravel('C'):
        o.write(s + str(v))
        s = ","
        i += 1
        if i % 10 == 0:
            print >>o, ","
            s = "  "
    print >>o, "};"
    return a.size * np.int32().nbytes

def print_maptype_array(o, name, a, prefix):
    print >>o, "%smaptype %s[%d] = {" % (prefix, name, a.size)
    s = "  "
    i = 0
    for v in a.ravel('C'):
        o.write(s + str(v))
        s = ","
        i += 1
        if i % 10 == 0:
            print >>o, ","
            s = "  "
    print >>o, "};"
    align = 4
    return ((a.size + align - 1) // align) * align

def print_intpp_data(o, tag, prefix, func, rgdata, basegrid, Spline=TensorSpline):
    if prefix:
        prefix += " "
    sz = 0
    int_sz = np.int32().nbytes
    float_sz = np.float32().nbytes
    #ranges = []
    #orderlist = []
    #for r, k in rgdata:
    #    ranges.append(r)
    #    orderlist.append(k)
    spl = Spline(func, rgdata, basegrid)
    coeffs = spl.calc_coeffs()
    max_idx = 0
    for iv, val in enumerate(spl.knot_data):
        for v in val:
            if not v.used():
                continue
            max_idx = max(max_idx, v.mapping[-1])
    print >>o, "typedef %s maptype;" % Spline.max_idx_to_maptype(max_idx)
    n = 0
    for iv, val in enumerate(spl.knot_data):
        for v in val:
            if v.used():
                n += 1
                v.cut_mapping()
        print >>o, "%sreal x0%s_%d[%d] = {%s};" % (prefix, tag, iv, n, ", ".join([str(float(v.get_bbox()[0])) for v in val if v.used()]))
        print >>o, "%sreal xe%s_%d[%d] = {%s};" % (prefix, tag, iv, n, ", ".join([str(float(v.get_bbox()[1])) for v in val if v.used()]))
        print >>o, "%sreal hi%s_%d[%d] = {%s};" % (prefix, tag, iv, n, ", ".join([str(v.inv_h()) for v in val if v.used()]))
        print >>o, "%sint k%s_%d[%d] = {%s};" % (prefix, tag, iv, n, ", ".join([str(v.get_order()) for v in val if v.used()]))
        print >>o, "%sint nmap%s_%d[%d] = {%s};" % (prefix, tag, iv, n, ", ".join([str(v.count()) for v in val if v.used()]))
        print >>o, "%sint n%s_%d[%d] = {%s};" % (prefix, tag, iv, n, ", ".join([str(v.num_coeffs()) for v in val if v.used()]))
        sz += (2*int_sz+2*float_sz) * n;
        #grd = np.mgrid[ranges]
        #grd_shape = grd.shape
        #numpoints = np.product(grd_shape[1:])
        #grd = grd.reshape(grd_shape[0], numpoints, order='F')
        #grd = grd.T
        #fnc = np.array(func(grd, True).T, order='C')
        #fnc.shape = (nvals,)+tuple(reversed(grd_shape[1:]))
        #fnc = fnc.T
        #args = [np.arange(s.step.imag) for s in ranges]
        ti = []
        ai = []
        ci = []
        #idx = [slice(None)]*n+[0]
        #fnca = []
        #for i in range(nvals):
        #    idx[n] = i
        #    fnca.append(fnc[idx])
        #tl, ca = b_ink(args, fnca, orderlist)
        for j in range(len(val)):
            if not val[j].used():
                continue
            t = "t%d_%d%s" % (j, iv, tag)
            ti.append(t)
            sz += print_float_array(o, t, val[j].knots, prefix)
            a = "a%d_%d%s" % (j, iv, tag)
            ai.append(a)
            sz += print_maptype_array(o, a, val[j].mapping, prefix)
        for i in range(1):
            t = "c%d_%d%s" % (i, iv, tag)
            ci.append(t)
            sz += print_float_array(o, t, coeffs[iv], prefix)
        print >>o, "%smaptype *map%s_%d[%d] = {%s};" % (prefix, tag, iv, n, ", ".join(ai))
        print >>o, "%s%s *t%s_%d[%d] = {%s};" % (prefix, dtp, tag, iv, n, ", ".join(ti))
        print >>o, "%s%s *c%s_%d[%d] = {%s};" % (prefix, dtp, tag, iv, 1, ", ".join(ci))
    sz += (n + len(basegrid)) * float_sz
    return sz, max_idx, spl

def print_header_file_start(h):
    print >>h, """\
#ifndef _DATA_H
#define _DATA_H 1

#ifndef NO_INTPP_INCLUDES
#include "intpp.h"
#endif

namespace AmpData {
    extern real b0;
    extern real b1;
    extern real a1;
    extern int fs;

"""

def print_header_file_end(h):
    print >>h, """\
}

#endif /* !_DATA_H */
"""

def print_header_file_entry(h, f):
    print >>h, "namespace %s { extern splinedata sd; }" % f

def print_header(o):
    print >>o, '#ifndef NO_INTPP_INCLUDES'
    print >>o, '#include "intpp.h"'
    print >>o, '#endif'
    print >>o, "namespace AmpData {"
    if 0:
        b0, b1, a1 = circuit.PhaseSplitter().feedback_coeff(1.0)
        print >>o, "real b0 = %g;" % b0
        print >>o, "real b1 = %g;" % b1
        print >>o, "real a1 = 1 - %g;" % (1 - a1)
    #print >>o, "int fs = %d;" % circuit.FS
    return 3 * np.float32().nbytes + np.int32().nbytes

def print_footer(o):
    print >>o, "} // namespace AmpData"
    return 0


class ArgumentError(ValueError):
    pass

class ValueGrid(object):

    def __init__(self, func, ranges, nvals):
        self.ranges = ranges
        self.grd = np.mgrid[ranges]
        self.values = dk_lib.calc_grid(func, self.grd, nvals)
        #print "%.2e" % np.max(self.values)

    def max_jacobi_row(self, vals):
        return np.array()

    def max_jacobi(self):
        n = self.values.shape[0]
        m = len(self.ranges)
        mj = np.zeros((n,m))
        for i in range(n):
            mj[i] = [np.max(abs(np.diff(self.values[i], axis=j))) for j in range(m)]
        for i in range(m):
            r = self.ranges[i]
            if r.start != r.stop:
                mj[:,i] /= (r.stop - r.start) / (r.step.imag-1)
            else:
                mj[:,i] = 0
        return mj

def estimate_max_jacobi(func, ranges, error, nvals):
    J = None
    n = np.ones(len(ranges), dtype=int) * 2
    while True:
        v = ValueGrid(func, [slice(r[0],r[1],i*1j) for i, r in zip(n, ranges)], nvals)
        J1 = v.max_jacobi()
        if J is not None:
            dv = np.max(abs(J1)+abs(J), axis=0)
            m = 2 * np.max(abs(J1-J), axis=0) / np.where(dv == 0, 1., dv)
            print "#%s: %s" % (n, m)
            if np.max(m) < error:
                #print "##", np.max(J1), np.max(v.values)
                return J1, v
            for i, e in enumerate(m):
                if e >= error:
                    n[i] *= 2
        else:
            n *= 2
        J = J1


class MyTensorSpline(TensorSpline):
    def __init__(self, func, ranges, basegrid):
        self.func = func
        self.ranges = ranges
        self.basegrid = basegrid
        self.knot_data = np.empty((len([b for b in basegrid if b[0] is not None]), len(ranges)), dtype=object)
        self.logger = logging.getLogger("approx")
        #bg = []
        self.coeffs = []
        self.axes = []
        kd_idx = 0
        for i_fnc, (rng, pre, post, err, opt) in enumerate(basegrid):
            if rng is None:
                continue
            grd, fnc, axes, axgrids = self.table_approximation(err, i_fnc, rng)
            grd, fnc, axes = self.trim_table(grd, fnc, axes, err, self.func.i0[i_fnc])
            self.logger.info("%d, %s, %s" % (i_fnc, fnc.shape, axgrids))
            #print fnc.ptp(), grd.shape, fnc.shape, len(axes), axgrids
            #grd.dump("grd.data")
            #fnc.dump("fnc.data")
            #import pickle
            #pickle.dump(axes, file("axes.data","w"))
            #raise SystemExit
            self.coeffs.append(fnc)
            self.axes.append(axes)
            for i, (ax, ag, cg, r) in enumerate(zip(axes, axgrids, func.basegrid[i_fnc][0], rng)):
                order_type = r[1]
                w = ax[-1]-ax[0]
                if w == 0:
                    order_type = None
                if order_type is not None:
                    if np.amax(fnc.ptp(axis=i)) < 1e-6 * fnc.ptp():
                        self.logger.info(
                            "%s[%d,%d]: const: %g" %
                            (func.comp_id, i_fnc, i, np.amax(fnc.ptp(axis=i)) / fnc.ptp()))
                        order_type = None
                if order_type is not None:
                    if isinstance(order_type, int):
                        tp = 's'
                        order = order_type
                    else:
                        tp = order_type[0]
                        order = order_type[1]
                    idx = np.array(np.rint((ax-ax[0])/w * ag), dtype=int)
                    if order > len(ax):
                        order = 2
                    self.knot_data[kd_idx, i] = self.mk_result(idx, ax, order, tp, slice(ax[0], ax[-1], (ag+1)*1j))
                else:
                    self.knot_data[kd_idx, i] = KnotData(None,None,slice(ax[0], ax[-1], 1j),None)
            kd_idx += 1

    @staticmethod
    def fromspline(xk, cvals, order):
        xk2 = xk[order:-order]
        N = len(xk2)-1
        sivals = np.empty((order+1, N), dtype=float)
        for m in xrange(order, -1, -1):
            fact = np.product(np.arange(m)+1)
            res = splev(xk2[:-1], (xk, cvals, order), m)
            res /= fact
            sivals[order-m, :] = res
        return sivals

    def calc_coeffs(self):
        coeffs = TensorSpline.calc_coeffs(self)
        cf = []
        for c, k in zip(coeffs, self.knot_data):
            k = [v for v in k if v.used()]
            if len(k) == 1 and k[0].tp == 'pp' and k[0].get_order() == 4:
                cs = np.squeeze(c)
                assert len(cs.shape) == 1
                cs = self.fromspline(k[0].knots, cs, 3).T
                c = cs.reshape([len(cs)*4 if s > 1 else 1 for s in c.shape])
            elif k[0].tp == 'h':
                k[0].tp = 'pp'
            else:
                # no other orders implemented
                k[0].tp = 's'
            cf.append(c)
        return cf

    def mk_result(self, idx, kn, k, tp, r):
        if k is None:
            return KnotData(None,None,slice(r.start, r.stop, 1j),None)
        f = reduce(fractions.gcd, idx)
        idx /= f
        a = np.empty(idx[-1], dtype=np.int32)
        for m, (i, j) in enumerate(itertools.izip_longest(idx[:-1], idx[1:])):
            a[i:j] = m
        a += k-1
        return KnotData(np.pad(kn, k-1, 'edge'), a, slice(r.start, r.stop, 1+(r.step-1)/f), (tp, k))

    def trim_axis_edge(self, i, idx, grd, fnc, axes, prec, i0):
        a = axes[i]
        if len(a) <= 2:
            return grd, fnc, axes
        if idx < 0:
            idx += len(a)
        s1 = [slice(None)] * len(axes)
        s2 = [slice(None)] * len(axes)
        s1[i] = idx
        fv = fnc[s1]
        s1[i] = idx-1
        s2[i] = idx+1
        fi = (fnc[s1] + fnc[s2]) * 0.5
        if callable(prec):
            df = prec(fi, fv, i0).any()
        else:
            df = (abs(fi - fv) > prec).any()
        if not df:
            sl = range(idx) + range(idx+1, len(a))
            axes[i] = axes[i][sl]
            s1[i] = sl
            fnc = fnc[s1]
            grd = grd[[slice(None)]+s1]
        return grd, fnc, axes
        
    def trim_axis(self, i, idx, grd, fnc, axes, prec, i0):
        a = axes[i]
        if len(a) <= 2:
            return grd, fnc, axes
        s = [slice(None)] * len(axes)
        s[i] = slice(idx, -1, 2)
        fi = fnc[s]
        s[i] = slice(idx+1, None, 2)
        fi = (fi + fnc[s]) * 0.5
        fsl = slice(idx+1, None, 2)
        s[i] = fsl
        fv = fnc[s]
        if callable(prec):
            df = prec(fi, fv, i0).any(axis=tuple([k for k in range(len(axes)) if k != i]))
        else:
            df = (abs(fi - fv) > prec).any(axis=tuple([k for k in range(len(axes)) if k != i]))
        if not df.all():
            sl = np.ones(len(a), dtype=bool)
            sl[fsl] = df
            axes[i] = axes[i][sl]
            s[i] = sl
            fnc = fnc[s]
            grd = grd[[slice(None)]+s]
        return grd, fnc, axes
        
    def trim_table(self, grd, fnc, axes, prec, i0):
        for i in range(len(axes)):
            grd, fnc, axes = self.trim_axis_edge(i, 1, grd, fnc, axes, prec, i0)
            grd, fnc, axes = self.trim_axis_edge(i, -2, grd, fnc, axes, prec, i0)
            grd, fnc, axes = self.trim_axis(i, 0, grd, fnc, axes, prec, i0)
            grd, fnc, axes = self.trim_axis(i, 1, grd, fnc, axes, prec, i0)
        return grd, fnc, axes

    def approx_one_axis(self, i, n, axes, s1, s2, fnc, ncalc_grid, prec, i0):
        a = axes[i]
        if len(a) == 1:
            return 0, None, None, None, None
        ax2 = (a[:-1]+a[1:]) * 0.5
        axeslist = axes[:i] + [ax2] + axes[i+1:]
        grd2 = dk_lib.mkgrid(axeslist)
        fnc2 = ncalc_grid(grd2)
        s1[i] = slice(None, -1)
        s2[i] = slice(1, None)
        fnc_intp = (fnc[s1] + fnc[s2]) * 0.5
        if callable(prec):
            df = prec(fnc_intp, fnc2, i0).any(axis=tuple([k for k in range(n) if k != i]))
        else:
            df = (abs(fnc_intp - fnc2) > prec).any(axis=tuple([k for k in range(n) if k != i]))
        k = np.count_nonzero(df)
        if k == 0:
            s1[i] = slice(None)
            s2[i] = slice(None)
        return k, df, ax2, grd2, fnc2

    def table_approximation(self, prec, i_fnc, rng):
        #print rng, prec
        def ncalc_grid(grd):
            return self.calc_grid(grd, None, None)[i_fnc]
        ranges = self.ranges
        n = len(ranges)
        axes = []
        axgrids = np.ones(n)
        start_size = 2
        for i, (r, g) in enumerate(zip(ranges, rng)):
            if g[1] is None:
                axes.append(np.array(((r[0]+r[1])*0.5,), dtype=np.float64))
            else:
                axes.append(np.linspace(r[0], r[1], start_size+1))
                axgrids[i] = start_size
        grd = dk_lib.mkgrid(axes)
        fnc = ncalc_grid(grd)
        s1 = [slice(None)]*n
        s2 = [slice(None)]*n
        g1 = [slice(None)]*(n+1)
        g2 = [slice(None)]*(n+1)
        nn = 0
        while True:
            nn += 1 
            inserted = False
            for i in range(len(axes)):
                k, df, ax2, grd2, fnc2 = self.approx_one_axis(i, n, axes, s1, s2, fnc, ncalc_grid, prec, self.func.i0[i_fnc])
                if k == 0:
                    continue
                axgrids[i] *= 2
                a = axes[i]
                inserted = True
                newshape = list(grd.shape)
                newshape[i+1] += k
                newgrd = np.empty(newshape)
                newshape = list(fnc.shape)
                newshape[i] += k
                newfnc = np.empty(newshape)
                newax = np.empty(len(a)+k)
                j = 0
                for k in range(len(a)):
                    g1[i+1] = j
                    g2[i+1] = k
                    newgrd[g1] = grd[g2]
                    s1[i] = j
                    s2[i] = k
                    newfnc[s1] = fnc[s2]
                    newax[j] = a[k]
                    j += 1
                    if k < len(df) and df[k]:
                        g1[i+1] = j
                        newgrd[g1] = grd2[g2]
                        s1[i] = j
                        newfnc[s1] = fnc2[s2]
                        newax[j] = ax2[k]
                        j += 1
                s1[i] = slice(None)
                s2[i] = slice(None)
                g1[i+1] = slice(None)
                g2[i+1] = slice(None)
                axes[i] = newax
                fnc = newfnc
                grd = newgrd
            self.logger.debug("%d, %s, %s" % (i_fnc, fnc.shape, axgrids))
            if not inserted:
                return grd, fnc, axes, axgrids


class TableGenerator(object):

    def __init__(self, v, args, name, param):
        parser = dk_simulator.Parser(v.S, v.V, v.FS, not args.backward_euler)
        if args.c_debug_load:
            raise NotImplemented("option c-debug-load with table generation")
        else:
            sim = dk_simulator.SimulatePy(dk_simulator.EquationSystem(parser), v.solver)
            cmod = dk_simulator.BuildCModule(
                name, sim, c_tempdir=args.c_tempdir, c_verbose=args.c_verbose,
                linearize=args.linearize, c_real="double")
            eq = cmod.eq
            p = cmod.get_executor()
        a = v.op_signal(timespan=1.02, op=parser.op)
        smpl = lambda tm: int(round(tm*v.FS))
        a[:,0] += param.input_signal.amplitude * dk_lib.genlogsweep(
            param.input_signal.startfreq, param.input_signal.stopfreq, v.FS,
            smpl(param.input_signal.pre), smpl(param.input_signal.timespan),
            smpl(param.input_signal.post))[0]
        ptp = p(a).ptp()
        print "ptp =", ptp
        print "nonlin function: OP value, range:"
        print np.column_stack((p.p0, p.minmax))

        nvals = p.nno
        J, vals = estimate_max_jacobi(p.nonlin, p.minmax, param.jacobi_estimate_error, nvals)
        J = np.matrix(J)
        dv = np.amax(np.append(abs(eq.Fo), abs(eq.Fo)*J*abs(eq.G0)*abs(eq.Co), axis=0), axis=0).A
        E = param.maxerr * ptp / np.where(dv == 0, 1e-20, dv)
        print "function error limits for max out error %g: %s" % (param.maxerr, ", ".join(["%.2g" % vv for vv in E.T]))

        grd_shape = vals.grd.shape
        numpoints = np.product(grd_shape[1:])
        grd = vals.grd.reshape(grd_shape[0], numpoints)
        fnc = vals.values.reshape(nvals, numpoints)
        with dk_lib.printoptions(precision=2, linewidth=200):
            print "covariance matrix (rows: variables, columns: functions):"
            print np.cov(grd, fnc)[:len(grd),len(grd):]

        o = StringIO()
        inst = StringIO()
        h = StringIO()
        class Comp:
            comp_id = name
            comp_name = name
            ranges = p.minmax
            basegrid = [
                [((1024, 2),), None, None, -E.T[0], True],
                ]
            # basegrid = [[((64, 2),(8, 2),(16, 2),(32, 2)), None, None, -E.T[0], True],
            #             [((280, 2),(8, 2),(16, 2),(32, 2)), None, None, -E.T[1], True],
            #             [((64, 2),(8, 2),(16, 2),(32, 2)), None, None, -E.T[2], True],
            #             [((64, 2),(8, 2),(16, 2),(32, 2)), None, None, -E.T[3], True],
            #             ]
            NVALS = nvals
            N_IN = grd_shape[0]
            NDIM = grd_shape[0]
            @staticmethod
            def __call__(v, with_state):
                return p.nonlin(v)
        if 1:
            maptype = self.write_files(Comp(), o, inst, h)
            c_debug_load = None
        else:
            c_debug_load = "gencode/dk_sim_0.so"
        extra_sources = dict(data_c=o.getvalue(), data_h=h.getvalue(), intpp_inst=inst.getvalue())
        sim = dk_simulator.SimulatePy(dk_simulator.EquationSystem(parser), v.solver)
        cmodt = dk_simulator.BuildCModule(
            name+"_table", sim, dict(method="table",name=name,maptype=maptype), extra_sources=extra_sources,
            c_tempdir="gencode", c_verbose=args.c_verbose, linearize=args.linearize, c_real=("float" if args.c_float else "double"))
        pt = cmodt.get_executor()
        self.p = p
        self.pt = pt
        ##
        rng = []
        for r, b in zip(Comp.ranges, Comp.basegrid[0][0]):
            off = 0.5 * (r[1]-r[0]) / (b[0]-1)
            rng.append(slice(r[0]+off, r[1]-off, (b[0]-1)*1j))
        v1 = ValueGrid(p.nonlin, rng, nvals)
        v2 = ValueGrid(pt.nonlin, rng, nvals)
        for i in range(len(Comp.basegrid)):
            print np.max(abs(v1.values[i]-v2.values[i]))

    @staticmethod
    def max_idx_to_maptype(max_idx):
        if max_idx >= 2**16:
            return "int"
        elif max_idx >= 2**8:
            return "unsigned short"
        else:
            return "unsigned char"

    @staticmethod
    def print_intpp_data(p):
        o = StringIO()
        print >>o, "namespace %s {" % p.comp_id
        r, max_idx, spl = print_intpp_data(o, "", "", p, p.ranges, p.basegrid, MyTensorSpline)
        maptype = TableGenerator.max_idx_to_maptype(max_idx)
        print >>o, "splinecoeffs<maptype> sc[%d] = {" % p.NVALS
        f_set = set()
        for j, kn in enumerate(spl.knot_data):
            fu = "splev"
            if kn[0].tp == 'pp':
                fu = "splev_pp"
            inst = "splinedata<%s>::%s<%s>" % (maptype, fu, ",".join([str(v.get_order()) for v in kn if v.used()]))
            f_set.add((inst,maptype))
            print >>o, "\t{x0_%d, xe_%d, hi_%d, k_%d, n_%d, nmap_%d, map_%d, t_%d, c_%d, %s}," % (j, j, j, j, j, j, j, j, j, inst)
        print >>o, "};"
        print >>o, "splinedata<maptype> sd = {"
        print >>o, "\tsc,"
        print >>o, "\t%d, /* number of calculated values */" % p.NVALS
        print >>o, "\t%d, /* number of input values */" % p.N_IN
        print >>o, "\t%d, /* number of output values */" % (p.NVALS-(p.NDIM-p.N_IN))
        print >>o, "\t%d, /* number of state values */" % (p.NDIM-p.N_IN)
        print >>o, '\t"%s",' % p.comp_id
        print >>o, "};"
        print >>o, "}; /* ! namespace %s */" % p.comp_id
        o.seek(0)
        return r, o.read(), f_set, p.comp_name, p.comp_id, max_idx, spl

    @staticmethod
    def write_files(comp, o, inst, h):
        procs = [TableGenerator.print_intpp_data(comp)]
        max_idx_all = 0
        for p in procs:
            s, f, i, comp_name, comp_id, max_idx, spl = p
            max_idx_all = max(max_idx_all, max_idx)
        #maptype = max_idx_to_maptype(max_idx_all)
        #o.write("typedef %s maptype;\n" % maptype)
        print_header_file_start(h)
        sz = print_header(o)
        l = []
        templ = set()
        for p in procs:
            s, f, i, comp_name, comp_id, max_idx, spl = p
            o.write(f)
            templ |= i
            l.append("%s: %d bytes" % (comp_name, s))
            sz += s
            print_header_file_entry(h, comp_id)
        sz += print_footer(o)
        l.append("data size sum: %d bytes" % sz)
        print >>o, "".join(["\n// " + s for s in l])
        print_header_file_end(h)
        maptype = "unsigned char"
        for v, maptype in sorted(templ):
            print >>inst, "template int %s(splinecoeffs<%s> *p, real xi[2], real *res);" % (v, maptype)
        return maptype, spl


class LoadedSchema(circ.Test):
    def __init__(self, params):
        v = vars(models)
        v["Tubes"] = circ.Tubes
        v["math"] = math
        if hasattr(params, "load_schema"):
            import mk_netlist
            exec mk_netlist.read_netlist(params.load_schema) in v
        else:
            with open(params.load_netlist) as f:
                exec f in v
        self.S = v["S"]
        self.V = v["V"]
        self.V["OP"] = getattr(params, "OP", [0.])
        if hasattr(params, "test_signal"):
            if hasattr(params.test_signal, "timespan"):
                self.timespan = params.test_signal.timespan
            def signal():
                a = self.op_signal()
                a[:,0] += params.test_signal.amplitude * self.sine_signal(params.test_signal.freq)[:,0]
                return a
            setattr(self, "signal", signal)


def generate_faust_module(plugindef, b, a, potlist, flt, pre_filter=None, build_script=None):
    import dk_templates
    d = {}
    if not potlist:
        d['have_master_slider'] = False
    else:
        d['have_master_slider'] = True
        d['master_slider_id'] = potlist[0][0]
    d['knob_ids'] = [t[0] for t in potlist]
    if not pre_filter == "dry_wet":
        ui = dk_templates.module_ui_template.render(d)
    else:
        ui = dk_templates.module_ui_dry_wet_template.render(d)
    d = {}
    d['plugindef'] = plugindef
    d['build_script'] = build_script
    d['sliders'] = [dict(id=t[0], name=t[1], loga=t[2], inv=t[3]) for t in potlist]
    d['pre_filter'] = '_' if pre_filter is None or "dry_wet" else pre_filter
    d['b_list'] = ",".join(["b%d/a0" % i for i in range(len(b))])
    d['a_list'] = ",".join(["a%d/a0" % i for i in range(1,len(a))])
    d['coeffs'] = "\n\n    ".join(flt.coeffs_as_faust_code('b', b) + flt.coeffs_as_faust_code('a', a))
    if not pre_filter == "dry_wet":
        dsp = dk_templates.faust_filter_template.render(d)
    else:
        dsp = dk_templates.faust_filter_dry_wet_template.render(d)
    return dsp, ui

def build_faust_module(plugindef, b, a, potlist, flt, datatype="float", pre_filter=None, build_script=None):
    dsp, ui = generate_faust_module(plugindef, b, a, potlist, flt, pre_filter, build_script)
    modname = plugindef.id
    dspname = "/tmp/%s.dsp" % modname
    uiname = "/tmp/%s_ui.cc" % modname
    with open(dspname,"w") as f:
        f.write(dsp)
    with open(uiname,"w") as f:
        f.write(ui)
    pgm = os.path.abspath("../../build-faust")
    opts = "-s" if datatype == "float" else ""
    os.system("%s %s -c %s" % (pgm, opts, dspname))

def get_circuit_instance(g, tests, args):
    if args.schema:
        class params(object):
            load_schema = args.schema
        t = os.path.splitext(os.path.basename(args.schema))[0]
        v = LoadedSchema(params)
        return v, t
    elif args.netlist:
        class params(object):
            load_netlist = args.netlist
        t = os.path.splitext(os.path.basename(args.netlist))[0]
        v = LoadedSchema(params)
        return v, t
    else:
        t = tests[0]
        return g[t](), t

def create_filter(g, tests, args):
    v, t = get_circuit_instance(g, tests, args)
    p = dk_simulator.Parser(v.S, v.V, v.FS, not args.backward_euler, create_filter=True, symbolic=args.filter_symbolic)
    if len(p.get_nonlin_funcs()) > 0:
        if args.filter_symbolic:
            raise ArgumentError("ciruit is nonlinear: symbolic formula generation not supported")
        p1 = dk_simulator.Parser(v.S, v.V, v.FS, not args.backward_euler)
        sim = dk_simulator.SimulatePy(dk_simulator.EquationSystem(p1), v.solver)
        J = sim.jacobi()
    else:
        J = None
    f = dk_simulator.LinearFilter(p, J)
    if args.filter_symbolic:
        if args.filter_s_coeffs:
            b, a, terms = f.get_s_coeffs()
            f.print_coeffs('b', b)
            f.print_coeffs('a', a)
            print "\nH = %s;" % terms
        else:
            b, a, terms = f.get_s_coeffs()
            f.print_coeffs('b', b)
            f.print_coeffs('a', a)
            B, A, c = f.transform_bilinear(terms)
            print "\nc = %s;" % c
            f.print_coeffs('B', B)
            f.print_coeffs('A', A)
    else:
        if args.filter_variable is None:
            if args.plot_spectrum:
                svar = f.convert_variable_dict({})
            else:
                svar = None
        else:
            try:
                svar = f.convert_variable_dict(dict([(s, float(vv)) for s, vv in [par.split("=") for par in args.filter_variable if par]]))
            except ValueError as e:
                raise ArgumentError(e)
        if args.filter_samplerate is None:
            if args.plot_spectrum:
                samplerate = 48000
            else:
                samplerate = None
        else:
            samplerate = args.filter_samplerate
        if args.plot_spectrum and args.plot_variable:
            pvar = [k for k in svar if args.plot_variable == str(k)]
            if not pvar:
                raise ArgumentError("variable %s not found" % args.plot_variable)
            pvar = pvar[0]
            del svar[pvar]
        b, a = f.get_z_coeffs(samplerate=samplerate, subst_var=svar)
        if args.plot_spectrum:
            if args.plot_variable:
                for e in p.element_name["P"]:
                    t = v.V[e[0]]
                    var = None
                    if isinstance(t, dict):
                        var = t.get("var")
                    if var is None:
                        var = "%sv" % e[0]
                    if var == args.plot_variable:
                        break
                else:
                    raise ArgumentError("variable %s not found" % args.plot_variable)
                if not isinstance(t, dict):
                    t = dict(value=t)
                loga = t.get('a', 0)
                inv = t.get('inv', 0)
                for i in range(5):
                    pot = i/4
                    lbl = "%s" % pot
                    if inv:
                        pot = 1 - pot
                    if loga:
                        pot = (math.exp(loga * pot) - 1) / (math.exp(loga) - 1)
                    w, h = f.spectrum([j.subs(pvar, pot) for j in b], [j.subs(pvar, pot) for j in a], 20, 10000)
                    pl.semilogx(w, np.where(h < -60, np.nan, h), label=lbl)
                pl.legend(loc='upper left')
            else:
                w, h = f.spectrum(b, a, 20, 10000)
                pl.semilogx(w, np.where(h < -60, np.nan, h))
            pl.grid()
            pl.show()
        elif args.create_module:
            l = []
            for e in set([e[0] for e in p.element_name["P"]]):
                t = v.V[e]
                if not isinstance(t, dict):
                    t = dict(value=t)
                var = t.get('var')
                if var is None:
                    var = str(e)+"v"
                name = t.get('name', var)
                loga = t.get('a', 0)
                inv = t.get('inv', 0)
                l.append((var, name, loga, inv))
            build_faust_module(dk_simulator.PluginDef(args.create_module), b, a, l, f)
        else:
            f.print_coeffs('b', b)
            f.print_coeffs('a', a)

def is_test(v):
    return isinstance(v, type) and issubclass(v, circ.Test) and v is not circ.Test

def plot_one(v, args, t):
    parser = dk_simulator.Parser(v.S, v.V, v.FS, not args.backward_euler)
    p = dk_simulator.get_executor(
        t, parser, v.solver, args.pure_python, c_tempdir=args.c_tempdir, c_verbose=args.c_verbose,
        c_debug_load=args.c_debug_load, linearize=args.linearize, c_real=("float" if args.c_float else "double"))
    if args.plot_spectrum:
        v.plot_spectrum(p, args.plot_variable)
    else:
        v.plot(p)
        if 0:
            for i, (p0, (s, e)) in enumerate(zip(p.p0, p.minmax)):
                print "%d: %g [%g .. %g]" % (i, p0, s, e)

def plot_output(g, tests, args):
    if args.schema or args.netlist:
        v, t = get_circuit_instance(g, tests, args)
        plot_one(v, args, t)
        return
    if not tests:
        testlist = [k for k, v in g.items() if is_test(v)]
        testlist.sort()
        for i, k in enumerate(testlist):
            if k.endswith("_test"):
                k = k[:-5]
            print "%2d: %s" % (i, k)
        print
        try:
            k = testlist[int(raw_input("Please select: "))]
        except (ValueError, KeyError):
            print "not found"
            raise SystemExit, 1
        except KeyboardInterrupt:
            print
            raise SystemExit, 1
        tests = [k]
    for t in tests:
        plot_one(g[t](), args, t)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('test', nargs="*",
                        help='name of test')
    parser.add_argument('-c', '--check', action='store_true',
                        help='check results of tests')
    parser.add_argument('-b', '--backward-euler', action='store_true',
                        help='use Backward-Euler integration instead of trapezoidal')
    parser.add_argument('-p', '--pure-python', action='store_true',
                        help='do not generate C code to speed up calculations')
    parser.add_argument('-f', '--func',
                        help='function table generation')
    parser.add_argument('-s', '--show', action='store_true',
                        help='plot nonlinear function')
    parser.add_argument('-S', '--schema',
                        help='use gschem .sch file as input')
    parser.add_argument('-N', '--netlist',
                        help='use netlist file as input (TBD: document file format)')
    parser.add_argument('--linearize', action='store_true',
                        help='build a linear small signal model')
    parser.add_argument('-y', '--filter', action='store_true',
                        help='calculate coefficients for linear filter (small signal model for nonlinear circuits)')
    parser.add_argument('--plot-spectrum', action='store_true',
                        help='plot spectrum of curcuit')
    parser.add_argument('--plot-variable',
                        help='display a set of curves by sweeping this the pot position for this variable from 0 to 1')
    parser.add_argument('--filter-s-coeffs', action='store_true',
                        help='calculate s coefficients (Laplace tranform for analog circuit)')
    parser.add_argument('--filter-symbolic', action='store_true',
                        help="use component symbols, don't replace by values (only valid for linear circuits)")
    parser.add_argument('--filter-samplerate', type=float,
                        help='sample rate for calculation of filter z coefficients (with this option the symbol fs will be used)')
    parser.add_argument('--filter-variable', action='append',
                        help='set a potentiometer variable in the form name=value (can be specified multiple times)')
    parser.add_argument('--create-module',
                        help='create a loadable Guitarix module (use with --filter)')
    parser.add_argument('--print-result', action='store_true',
                        help='print result (to be used as reference)')
    parser.add_argument('--c-verbose', action='store_true',
                        help='show compiler messages when generating internal module')
    parser.add_argument('--c-tempdir',
                        help='temp dir for module generation; not removed when finished')
    parser.add_argument('--c-float', action='store_true',
                        help='use float instead of double in c module')
    parser.add_argument('--c-debug-load',
                        help='load module (generated with --c-tempdir) instead of generating a new one')
    parser.add_argument('--c-samplerate', type=float,
                        help='sample rate for calculation of c module)')

    args = parser.parse_args()
    g = vars(circ)
    tests = []
    for t in args.test:
        if not (t in g and is_test(g[t])):
            tt = t + "_test"
            if not (tt in g and is_test(g[tt])):
                parser.error("%s is not a test" % t)
            t = tt
        tests.append(t)
    if args.check:
        if not tests:
            tests = [k for k, v in g.items() if is_test(v) and hasattr(v, 'result')]
            tests.sort()
        tn = [t[:-5] if t.endswith("_test") else t for t in tests]
        mlen = reduce(max, [len(t) for t in tn], 0)
        for t, nm in zip(tests, tn):
            sys.stdout.write("%-*s: " % (mlen, nm))
            sys.stdout.flush()
            v = g[t]()
            sys.stdout.write("%s\n" % v.check(t, args))
    elif args.func:
        t = os.path.splitext(os.path.basename(args.func))[0]
        class sweep(object):
            def __init__(self, startfreq=30, stopfreq=20000, timespan=1, pre = 0.01, post = 0.01):
                for i in "startfreq", "stopfreq", "timespan", "pre", "post":
                    setattr(self, i, vars()[i])
                self.amplitude = 1
            def __rmul__(self, m):
                self.amplitude *= m
                return self
            __mul__ = __rmul__
        class sine_signal(object):
            def __init__(self, freq=200, timespan=0.1):
                self.freq = freq
                self.timespan=timespan
                self.amplitude = 1
            def __rmul__(self, m):
                self.amplitude *= m
                return self
            __mul__ = __rmul__

        d = dict(sweep=sweep, sine_signal=sine_signal, circ=circ)
        try:
            with open(args.func) as f:
                exec f in d
        except IOError:
            raise SystemExit("error: can't open %s" % args.func)
        class params(object):
            jacobi_estimate_error = 0.1
        for k, v in d.items():
            if k != 'sweep' and not k.startswith("__"):
                setattr(params, k, v)
        if hasattr(params, "load_schema") or hasattr(params, "load_netlist"):
            v = LoadedSchema(params)
        else:
            v = params.schema()
        tg = TableGenerator(v, args, t, params)
        if hasattr(v, "signal"):
            x = v.timeline()
            s = v.signal()
            tg.p.x = tg.p.x0
            tg.p.v0 = tg.p.v00
            y1 = tg.p(s)
            pl.plot(x, y1, label="solver")
            y2 = tg.pt(s)
            pl.plot(x, y2, label="table")
            #
            #parser = dk_simulator.Parser(v.S, v.V, v.FS, not args.backward_euler)
            #p2 = dk_simulator.get_executor(t+"_table", parser, v.solver, args.pure_python,
            #                               c_debug_load="gencode/dk_sim_1.so")
            #y3 = p2(s)
            #pl.plot(x, y3, label="reloaded")
            #print np.max(abs(y3 - y1))
            #
            pl.grid()
            pl.legend()
            pl.show()
    elif args.show:
        t = tests[0]
        v = g[t]()
        parser = dk_simulator.Parser(v.S, v.V, v.FS, not args.backward_euler)
        p = dk_simulator.get_executor(
            t, parser, v.solver, args.pure_python, c_tempdir=args.c_tempdir,
            c_verbose=args.c_verbose, c_debug_load=args.c_debug_load)
        a = v.op_signal(timespan=1.02)
        smpl = lambda tm: int(round(tm*v.FS))
        a[:,0] += 2 * dk_lib.genlogsweep(30, 20000, v.FS, smpl(0.01), smpl(1), smpl(0.01))[0]
        ptp = p(a).ptp()
        rng = p.minmax
        ifunc = 3
        j = 0
        i = 1
        for k in range(4):
            if j == k:
            #if j <= k:
            #if k <= j:
                continue
            x = np.linspace(rng[j][0],rng[j][1],100)
            z = np.linspace(rng[k][0],rng[k][1],20)
            a = [[p0] for p0 in p.p0]
            a[j] = x
            a[k] = z
            y = dk_lib.calc_grid(p.nonlin, dk_lib.mkgrid(a))[ifunc]
            s = [slice(None)] * len(a)
            for n in range(len(s)):
                if n not in (j,k):
                    s[n] = 0
            pl.subplot(3,1,i)
            i += 1
            t = y[s]
            if j > k:
                t = t.T
            else:
                t = t
            if 0:
                t = np.diff(t, 2, axis=0)
                x = x[:-2]
            if 0:
                t = np.diff(t, 1, axis=0)
                x = x[:-1]
            for ii, tt in enumerate(t.T):
                off = 0
                if j == 0 and k == 1:
                    off = -0.55 * z[ii]
                if j == 0 and k == 2:
                    off = 0.9 * z[ii]
                if j == 0 and k == 3:
                    off = -0.9 * z[ii]
                off = 0
                pl.plot(x+off, tt)
            t = correlate(t[:,1:],t[:,:1],'same')
            m = t.argmax(axis=0)
            #pl.plot(np.diff(m))
            idx = (m, range(len(m)))
            #pl.plot(x[m], t[idx])
            pl.title("%d, %d" % (j, k))
            pl.grid()
        pl.show()
    elif args.filter:
        try:
            create_filter(g, tests, args)
        except ArgumentError as e:
            parser.error(e)
    else:
        plot_output(g, tests, args)


if __name__ == "__main__":
    main()



#! /usr/bin/env python
# encoding: utf-8

import Options
import Task

def set_options(opt):
	opt.add_option('--with-gcov', action='store_true', help='Use gcov for profiling', default=False)

def setup(bld):
	for name in('cc','cxx'):
		if name not in Task.TaskBase.classes:
			continue
		task = Task.TaskBase.classes[name]
		line = task.run.code.replace('SRC','SRC[0].abspath(env)')
		(fun,v) = Task.compile_fun(name,line,False)
		fun.code = line
		task.run = fun

def detect(conf):
	conf.find_program('gcov')
	gcov_flags = '-fprofile-arcs -ftest-coverage'.split()
	conf.env['CXXFLAGS'] = conf.env['CFLAGS'] = conf.env['LINKFLAGS'] = gcov_flags



def add_zita_convolver(bld, uselib, sources, incl, base=".."):
    if bld.env['ZITA_CONVOLVER']:
        uselib.append('ZITA_CONVOLVER')
    elif bld.env['CONVOLVER_FFMPEG']:
        sources.append(base+'/zita-convolver-ffmpeg/zita-convolver.cc')
        incl.append(base+'/zita-convolver-ffmpeg');
        uselib += ['LIBAVCODEC','LIBAVUTIL']
    else:
        sources.append(base+'/zita-convolver/zita-convolver.cc')
        incl.append(base+'/zita-convolver');

def add_zita_resampler(bld, uselib, sources, incl, base=".."):
    if bld.env['ZITA_RESAMPLER']:
        uselib.append('ZITA_RESAMPLER')
    else:
        sources.append(base+'/zita-resampler-1.1.0/resampler.cc')
        sources.append(base+'/zita-resampler-1.1.0/resampler-table.cc')
        incl.append(base+'/zita-resampler-1.1.0')

def get_faust_args(bld):
    float_arg = ["-s","40000","--float"]
    double_arg = ["--double"]
    if bld.env['FAUST_DOUBLE']:
        arg = ["--double"]
    else:
        arg = ["--float"]
    if bld.env['FAUST_VECTORIZE']:
        float_arg.append('--vectorize')
        arg.append('--vectorize')
        double_arg.append('--vectorize')
    elif bld.env['FAUST_VECTORIZE_FLOAT']:
        float_arg.append('--vectorize')
        if not bld.env['FAUST_DOUBLE']:
            arg.append('--vectorize')
    add_args = bld.env['FAUST_OPTIONS']
    if add_args:
        add_args = "--add=%s" % add_args
        float_arg.append(add_args)
        double_arg.append(add_args)
        arg.append(add_args)
    return arg, float_arg, double_arg
